# Note: this entire module is compiled by cython with wraparound=False
# See meson.build for details

from __future__ import annotations

import abc
import enum
import itertools
from collections.abc import Iterator
from typing import TYPE_CHECKING, Any

import cython
import numpy as np
from cython import bint, ssize_t
from ndindex import ChunkSize, Slice, Tuple, ndindex
from ndindex.ellipsis import ellipsis
from numpy.typing import NDArray

from .cytools import (
    ceil_a_over_b,
    count2stop,
    np_hsize_t,
    smallest_step_after,
    stop2count,
)
from .tools import asarray

if TYPE_CHECKING:  # pragma: nocover
    # TODO import from typing and remove quotes (requires Python 3.10)
    # TODO use type <name> = ... (requires Python 3.12)
    from typing_extensions import TypeAlias

    AnySlicer: TypeAlias = "slice | NDArray[np_hsize_t] | int"
    AnySlicerND: TypeAlias = tuple[AnySlicer, ...]

if cython.compiled:  # pragma: nocover
    from cython.cimports.versioned_hdf5.cytools import (  # type: ignore
        ceil_a_over_b,
        count2stop,
        hsize_t,
        smallest_step_after,
        stop2count,
    )

if cython.compiled:  # pragma: nocover
    from cython.cimports.versioned_hdf5.cytools import (  # type: ignore
        ceil_a_over_b,
        smallest_step_after,
    )


class DropAxis(enum.Enum):
    _drop_axis = 0


# Returned instead of an AnySlicer. Signals that the axis should be removed when
# aggregated into an AnySlicerND.
DROP_AXIS = DropAxis._drop_axis


# This is an abstract class, which should inherit from abc.ABC
# or have metaclass=abc.ABCMeta. Neither are supported by Cython though.
@cython.cclass
class IndexChunkMapper:
    """Abstract class that manipulates a numpy fancy index along a single axis of a
    chunked array

    Parameters
    ----------
    chunk_indices:
        Array of indices of all the chunks involved in the selection along the axis
    chunk_size:
        Size of each chunk, in points, along the axis
    dset_size:
        Size of the whole array, in points, along the axis
    """

    chunk_indices: hsize_t[:]
    dset_size: hsize_t
    chunk_size: hsize_t
    n_chunks: hsize_t
    last_chunk_size: hsize_t

    def __init__(
        self,
        chunk_indices: hsize_t[:],
        dset_size: hsize_t,
        chunk_size: hsize_t,
    ):
        self.chunk_indices = chunk_indices
        self.dset_size = dset_size
        self.chunk_size = chunk_size
        self.n_chunks = ceil_a_over_b(dset_size, chunk_size)
        self.last_chunk_size = (dset_size % chunk_size) or chunk_size

    @cython.cfunc
    @cython.nogil
    @cython.exceptval(check=False)
    def _chunk_start_stop(self, chunk_idx: hsize_t) -> tuple[hsize_t, hsize_t]:
        """Return the range of points [a, b[ of the chunk indexed by chunk_idx"""
        if not cython.compiled:
            chunk_idx = int(chunk_idx)  # np.uint64 + int -> np.float64

        start = chunk_idx * self.chunk_size
        stop = min(start + self.chunk_size, self.dset_size)
        return start, stop

    @cython.ccall
    @abc.abstractmethod
    def chunk_submap(
        self, chunk_idx: hsize_t
    ) -> tuple[Slice, AnySlicer | DropAxis, AnySlicer]:
        """Given a chunk index, return a tuple of

        data_dict key
            key of the data_dict (see build_data_dict())
        value_subidx
            the slicer selecting the points within the sliced array
            (the return value for __getitem__, the value parameter for __setitem__)
        chunk_subidx
            the slicer selecting the points within the input chunks.

        In other words, in the simplified one-dimensional case:

            _, value_subidx, chunk_subidx = mapper.chunk_submap(i)
            chunk_view = base_arr[chunk_idx * chunk_size:(chunk_idx + 1) * chunk_size]
            return_value[value_subidx] = chunk_view[chunk_subidx]  # __getitem__
            chunk_view[chunk_subidx] = value_param[value_subidx]  # __setitem__
        """

    @cython.ccall
    @abc.abstractmethod
    def read_many_slices_params(
        self,
    ) -> tuple[NDArray[np_hsize_t], NDArray[np_hsize_t] | None]:
        """Return the parameters for read_many_slices() for all chunks of the selection.

        Returns
        -------
        Tuple of two 2D arrays, (slices, chunk_to_slices), or (slices, None).
        The presence of chunk_to_slices changes between basic and fancy indexes.

        **Basic indexing**

        For basic indexing and in simple special cases of fancy indexes, there is a
        1:1 correlation between selected chunks and slices to tranfer. In this

        len(slices) == len(self.chunk_indices).
        chunk_to_slices is None

        Each row represents a slice to transfer the data for the matching chunk on
        chunk_indices.

        To retrieve the absolute chunk index and the slice for the i-th chunk of the
        selection:

        >> chunk_idx = mapper.chunk_indices[i]
        >> slices, _ = mapper.read_many_slices_params()
        >> slice_i = slices[i, :]


        **Fancy indexing**

        Each chunk is transferred by 1 or more slices (1:N).

        len(slices) > len(self.chunk_indices)
        len(chunk_to_slices) == len(self.chunk_indices) + 1

        chunk_to_slices is a 1D array with as many points as chunk_indices plus one,
        with each point representing the start of the slices view for that chunk. The
        slices for a chunk end where the slices for the next chunk begin.

        The first and last point are there just for convenience; this is always true:
            chunk_to_slices[0] == 0
            chunk_to_slices[-1] == len(slices)

        To retrieve the absolute chunk index and the slices for the i-th chunk of the
        selection:

        >> chunk_idx = mapper.chunk_indices[i]
        >> slices, chunk_to_slices = mapper.read_many_slices_params()
        >> slices_i = slices[chunk_to_slices[i]:chunk_to_slices[i+1]]

        Note: in any n-dimensional index there is always at most one fancy index.


        **The slices array**

        The slices array features one row per slice and exactly five columns:

        chunk_sub_start
            The start of the slice, relative to the first point of the chunk
        value_sub_start
            The start of the slice, relative to the whole __getitem__ return value or
            the whole __setitem__ value parameter. 0 for scalar indices.
        count
            The number of points selected by the slice
        chunk_sub_stride
            How many points to skip within the chunks between each selected point,
            a.k.a. step
        value_sub_stride
            How many points to skip within the __getitem__ return value or the
            __setitem__ value parameter between each selected point.
            This is always 1 except for the case of out-of-order fancy indexes.

        See the enum subchunk_map.pxd::ReadManySlicesColumn.

        The mapping from the columns of the slices view to the parameters of
        read_many_slices() is as follows:

        __getitem__ (from slab to output value)
            src_start  = chunk_sub_start + (slab_offset if axis==0 else 0)
            dst_start  = value_sub_start
            count      = count
            src_stride = chunk_sub_stride
            dst_stride = value_sub_stride

        __setitem__ (from input value to slab)
            src_start  = value_sub_start
            dst_start  = chunk_sub_start + (slab_offset if axis==0 else 0)
            count      = count
            src_stride = value_sub_stride
            dst_stride = chunk_sub_stride

        slab-to-slab transfer
            src_start  = chunk_sub_start + (src_slab_offset if axis==0 else 0)
            dst_start  = chunk_sub_start + (dst_slab_offset if axis==0 else 0)
            count      = count
            src_stride = chunk_sub_stride
            dst_stride = chunk_sub_stride
        """

    @cython.ccall
    @abc.abstractmethod
    def chunks_indexer(self):  # -> slice | NDArray[np_hsize_t]:
        """Return a numpy basic or advanced index, to be applied along the matching axis
        to an array with one point per chunk, that returns all chunks involved in the
        selection without altering the shape of the array.
        """

    @cython.ccall
    @abc.abstractmethod
    def whole_chunks_idxidx(self):  # -> slice | NDArray[np.intp]:
        """Return an index that further refines the output of chunks_indexer and
        selects only the chunks where all points of the chunk are included in the
        selection.

        e.g. if the index of this mapper is [True, False, False, False, True, True]
        and self.chunk_size=2, then:

        - self.chunks_indexer() -> [0, 2]
        - self.whole_chunks_idxidx() -> [1]  # relative to the output of chunks_indexer

        In other words:

        >> all_chunk_indices = np.arange(self.n_chunks)
        >> selected_chunk_indices = all_chunk_indices[self.chunks_indexer()]
        >> whole_chunk_indices = selected_chunk_indices[self.whole_chunks_idxidx()]
        """

    @property
    def value_view_idx(self) -> slice | None:
        """Return an index, to be applied to the matching axis of the __getitem__
        return value or the __setitem__ value parameter, that will create a view with
        has the same dimensionality as the base array.

        This allows passing the view to read_many_slices().

        Example
        -------
        >>> idx = (slice(3), 1, slice(None))
        >>> chunks = (4, 4, 4)
        >>> shape = (10, 10, 10)
        >>> ndindex_idx, mappers = index_chunk_mappers(idx, shape, chunks)
        >>> value = np.empty(ndindex_idx.newshape(shape))  # e.g. for __getitem__
        >>> value_view = value[tuple(mapper.value_view_idx for mapper in mappers)]
        >>> value.shape
        (3, 10)
        >>> value_view.shape
        (3, 1, 10)
        >>> value_view.base is value
        True
        """
        return slice(None)


@cython.cclass
class BasicChunkMapper(IndexChunkMapper):
    """Abstract IndexChunkMapper for numpy basic indexing (slices and integers)"""

    @cython.ccall
    def read_many_slices_params(self) -> tuple[NDArray[np_hsize_t], None]:
        n_sel_chunks = len(self.chunk_indices)
        slices = np.empty((n_sel_chunks, 5), dtype=np_hsize_t)
        slices_v: hsize_t[:, :] = slices

        with cython.nogil:
            for idxidx in range(n_sel_chunks):
                chunk_idx = self.chunk_indices[idxidx]
                (
                    chunk_sub_start,
                    value_sub_start,
                    count,
                    chunk_sub_stride,
                ) = self._read_many_slices_param(chunk_idx)
                # For reference, see subchunk_map.pxd::ReadManySlicesColumn
                slices_v[idxidx, 0] = chunk_sub_start
                slices_v[idxidx, 1] = value_sub_start
                slices_v[idxidx, 2] = count
                slices_v[idxidx, 3] = chunk_sub_stride
                slices_v[idxidx, 4] = 1  # value_sub_stride

        return slices, None

    @cython.cfunc
    @cython.nogil
    @cython.exceptval(check=False)
    @abc.abstractmethod
    def _read_many_slices_param(
        self,
        chunk_idx: hsize_t,
    ) -> tuple[hsize_t, hsize_t, hsize_t, hsize_t]:
        """Return the parameters to read_many_slices() for a single chunk.

        Returns tuple of:
        - chunk_sub_start
        - value_sub_start
        - count
        - chunk_sub_stride
        """


@cython.cclass
class SliceMapper(BasicChunkMapper):
    """IndexChunkMapper for slices"""

    start: hsize_t = cython.declare("hsize_t", visibility="readonly")
    stop: hsize_t = cython.declare("hsize_t", visibility="readonly")
    step: hsize_t = cython.declare("hsize_t", visibility="readonly")

    def __init__(
        self,
        idx: slice,
        dset_size: hsize_t,
        chunk_size: hsize_t,
    ):
        if idx.step <= 0:
            raise NotImplementedError("Slice step must be positive")

        self.start = idx.start
        self.stop = idx.stop
        self.step = idx.step

        if self.step > chunk_size:
            n = (self.stop - self.start + self.step - 1) // self.step
            chunk_indices = (
                self.start + np.arange(n, dtype=np_hsize_t) * self.step
            ) // chunk_size
        else:
            chunk_start = self.start // chunk_size
            chunk_stop = (self.stop + chunk_size - 1) // chunk_size
            chunk_indices = np.arange(chunk_start, chunk_stop, dtype=np_hsize_t)

        super().__init__(chunk_indices, dset_size, chunk_size)

    @cython.ccall
    def chunk_submap(self, chunk_idx: hsize_t) -> tuple[Slice, slice, slice]:
        chunk_start, chunk_stop = self._chunk_start_stop(chunk_idx)
        sel_start = self.start
        sel_stop = self.stop
        sel_step = self.step

        abs_start = max(chunk_start, sel_start)
        # Get the smallest lcm multiple of common that is >= start
        abs_start = smallest_step_after(abs_start, sel_start % sel_step, sel_step)

        # shift start so that it is relative to index
        value_sub_start = (abs_start - sel_start) // sel_step
        value_sub_stop = ceil_a_over_b(min(chunk_stop, sel_stop) - sel_start, sel_step)

        chunk_sub_start = abs_start - chunk_start
        count = value_sub_stop - value_sub_start
        chunk_sub_stop = count2stop(chunk_sub_start, count, sel_step)

        return (
            Slice(chunk_start, chunk_stop, 1),
            slice(value_sub_start, value_sub_stop, 1),
            slice(chunk_sub_start, chunk_sub_stop, sel_step),
        )

    @cython.cfunc
    @cython.nogil
    @cython.exceptval(check=False)
    def _read_many_slices_param(
        self,
        chunk_idx: hsize_t,
    ) -> tuple[hsize_t, hsize_t, hsize_t, hsize_t]:
        chunk_start, chunk_stop = self._chunk_start_stop(chunk_idx)
        sel_start = self.start
        sel_stop = self.stop
        sel_step = self.step

        abs_start = max(chunk_start, sel_start)
        # Get the smallest lcm multiple of common that is >= start
        abs_start = smallest_step_after(abs_start, sel_start % sel_step, sel_step)

        # shift start so that it is relative to index
        value_sub_start = (abs_start - sel_start) // sel_step
        value_sub_stop = ceil_a_over_b(min(chunk_stop, sel_stop) - sel_start, sel_step)

        count = value_sub_stop - value_sub_start
        chunk_sub_start = abs_start - chunk_start
        chunk_sub_stride = sel_step

        return chunk_sub_start, value_sub_start, count, chunk_sub_stride

    @cython.ccall
    def chunks_indexer(self):  # -> slice | NDArray[np_hsize_t]:
        indices = self.chunk_indices
        if self.step > self.chunk_size:
            return _maybe_array_idx_to_slice(np.asarray(indices))
        else:
            idx_len = len(indices)
            assert idx_len > 0  # Empty slices are filtered out by index_chunk_mappers()
            chunk_start = indices[0]
            chunk_stop = indices[idx_len - 1] + 1
            return slice(int(chunk_start), int(chunk_stop), 1)

    @cython.ccall
    def whole_chunks_idxidx(self):
        indices = self.chunk_indices
        idx_len = len(indices)
        assert idx_len > 0  # Empty slices are filtered out by index_chunk_mappers()

        if self.chunk_size == 1:
            # All chunks are wholly selected
            return slice(0, idx_len, 1)

        if self.step > 1:
            if self.last_chunk_size == 1 and indices[idx_len - 1] == self.n_chunks - 1:
                # Last chunk contains exactly one point, so it's wholly covered.
                # For all other chunks, chunk_size > 1 and step > 1 so the selection
                # will never cover a whole chunk
                return slice(idx_len - 1, idx_len, 1)
            else:
                return slice(0, 0, 1)

        # step==1. The first and last chunk may be partially selected;
        # the rest are always wholly selected.
        idxidx_start = int(self.start % self.chunk_size != 0)
        idxidx_stop = idx_len
        if self.stop < self.dset_size and self.stop % self.chunk_size > 0:
            idxidx_stop -= 1
        return slice(idxidx_start, idxidx_stop, 1)


@cython.cclass
class IntegerMapper(BasicChunkMapper):
    """IndexChunkMapper for scalar integer indices"""

    idx: hsize_t = cython.declare("hsize_t", visibility="readonly")

    def __init__(self, idx: hsize_t, dset_size: hsize_t, chunk_size: hsize_t):
        assert 0 <= idx < dset_size
        self.idx = idx
        chunk_indices = np.array([idx // chunk_size], dtype=np_hsize_t)
        super().__init__(chunk_indices, dset_size, chunk_size)

    @cython.ccall
    def chunk_submap(self, chunk_idx: hsize_t) -> tuple[Slice, DropAxis, int]:
        chunk_start, chunk_stop = self._chunk_start_stop(chunk_idx)
        chunk_sub_idx = self.idx - chunk_start
        return Slice(chunk_start, chunk_stop, 1), DROP_AXIS, chunk_sub_idx

    @cython.cfunc
    @cython.nogil
    @cython.exceptval(check=False)
    def _read_many_slices_param(
        self,
        chunk_idx: hsize_t,
    ) -> tuple[hsize_t, hsize_t, hsize_t, hsize_t]:
        chunk_start = chunk_idx * self.chunk_size
        return (
            self.idx - chunk_start,  # chunk_sub_start
            0,  # value_sub_start
            1,  # count
            1,  # chunk_sub_stride
        )

    @cython.ccall
    def chunks_indexer(self):
        # a[i] would change the shape
        # a[[i]] (the default without overriding this method) would return a copy
        # a[i:i+1] returns a view, which is faster
        i = int(self.chunk_indices[0])
        return slice(i, i + 1, 1)

    @cython.ccall
    def whole_chunks_idxidx(self):
        if self.chunk_size == 1:
            return slice(0, 1, 1)

        # If the index is in the last chunk and the last chunk is size 1, then it's
        # wholly selected. In any other case, it's partially selected.
        partial = self.dset_size % self.chunk_size
        if partial != 1:
            return slice(0, 0, 1)
        if self.chunk_indices[0] != self.dset_size // self.chunk_size:
            return slice(0, 0, 1)
        return slice(0, 1, 1)

    @property
    def value_view_idx(self) -> slice | None:
        return None


@cython.cclass
class EntireChunksMapper(BasicChunkMapper):
    """Special mapper that selects all points on the chunks selected by another mapper.

    This is used to load the entire chunk for the purpose of caching when the
    actual selection may only target part of it.
    """

    _chunks_indexer: slice | NDArray[np.intp]

    def __init__(self, other: IndexChunkMapper):
        self._chunks_indexer = other.chunks_indexer()
        super().__init__(other.chunk_indices, other.dset_size, other.chunk_size)

    @cython.ccall
    def chunk_submap(
        self, chunk_idx: hsize_t
    ) -> tuple[Slice, AnySlicer | DropAxis, AnySlicer]:
        raise NotImplementedError(  # pragma: nocover
            "not used in legacy as_subchunk_map"
        )

    @cython.cfunc
    @cython.nogil
    @cython.exceptval(check=False)
    def _read_many_slices_param(
        self, chunk_idx: hsize_t
    ) -> tuple[hsize_t, hsize_t, hsize_t, hsize_t]:
        chunk_start, chunk_stop = self._chunk_start_stop(chunk_idx)

        return (
            0,  # chunk_sub_start
            chunk_start,  # value_sub_start
            chunk_stop - chunk_start,  # count
            1,  # chunk_sub_stride
        )

    @cython.ccall
    def chunks_indexer(self):
        return self._chunks_indexer

    @cython.ccall
    def whole_chunks_idxidx(self):
        return slice(0, len(self.chunk_indices), 1)


@cython.cclass
class IntegerArrayMapper(IndexChunkMapper):
    """IndexChunkMapper for one-dimensional fancy integer array indices.
    This is also used for boolean indices (preprocessed with np.flatnonzero()).
    """

    idx: NDArray[np_hsize_t] = cython.declare(object, visibility="readonly")
    is_ascending: bint = cython.declare(bint, visibility="readonly")

    def __init__(
        self,
        idx: NDArray[np_hsize_t],
        dset_size: hsize_t,
        chunk_size: hsize_t,
    ):
        self.idx = idx

        self.is_ascending = False
        idx_v: hsize_t[:] = idx
        prev = idx_v[0]
        for i in range(1, len(idx_v)):
            if idx_v[i] < prev:
                break
            prev = idx_v[i]
        else:
            self.is_ascending = True

        chunk_indices = np.unique(idx // chunk_size)
        super().__init__(chunk_indices, dset_size, chunk_size)

    @cython.ccall
    def chunk_submap(
        self, chunk_idx: hsize_t
    ) -> tuple[Slice, NDArray[np_hsize_t] | slice, NDArray[np_hsize_t] | slice]:
        chunk_start, chunk_stop = self._chunk_start_stop(chunk_idx)

        if self.is_ascending:
            # O(n*logn)
            start_idx, stop_idx = np.searchsorted(self.idx, [chunk_start, chunk_stop])
            mask = slice(int(start_idx), int(stop_idx), 1)
        # TODO optimize monotonic descending
        else:
            # O(n^2)
            mask = (chunk_start <= self.idx) & (self.idx < chunk_stop)
            mask = _maybe_array_idx_to_slice(mask)

        return (
            Slice(chunk_start, chunk_stop, 1),
            mask,
            _maybe_array_idx_to_slice(self.idx[mask] - chunk_start),
        )

    @cython.ccall
    def read_many_slices_params(
        self,
    ) -> tuple[NDArray[np_hsize_t], NDArray[np_hsize_t] | None]:
        n_sel_chunks = len(self.chunk_indices)
        max_rows = len(self.idx)

        slices = np.empty((max_rows, 5), dtype=np_hsize_t)
        slices_v: hsize_t[:, :] = slices

        # Each selected chunk containst at least one point by construction
        assert n_sel_chunks <= max_rows
        if n_sel_chunks == max_rows:
            # Because we know we are selecting at least one point per chunk of
            # chunk_indices, then if there are exactly as many chunks as there are
            # points we know that each point falls in a separate chunk and there will be
            # exactly one slice(idx[i], idx[i] + 1, 1) per chunk, so there is no need
            # for the 1:N mapping table chunk_to_slices.
            chunks_to_slices = None
        else:
            # At least one chunk contains more than one point
            chunks_to_slices = np.empty(n_sel_chunks + 1, dtype=np_hsize_t)
            chunk_to_slices_v: hsize_t[:] = chunks_to_slices

        if self.is_ascending:
            starts_stops = np.empty((n_sel_chunks, 2), dtype=np_hsize_t)
            starts_stops_v: hsize_t[:, :] = starts_stops
            for chunk_idxidx in range(n_sel_chunks):
                chunk_idx = self.chunk_indices[chunk_idxidx]
                chunk_start, chunk_stop = self._chunk_start_stop(chunk_idx)
                starts_stops_v[chunk_idxidx, 0] = chunk_start
                starts_stops_v[chunk_idxidx, 1] = chunk_stop

            # This should be faster than calling searchsorted n_sel_chunks times
            # TODO test performance and consider rewriting this in cython, with the
            # extra assumption that starts_stops is sorted,  which allows reducing
            # big-O complexity from O(n*logn) to O(n).
            starts_stops_idxidx: ssize_t[:, :] = np.searchsorted(self.idx, starts_stops)
            idx_v: hsize_t[:] = self.idx
            value_full_idx_v: hsize_t[:] = np.arange(max_rows, dtype=np_hsize_t)

        value_sub_idx: hsize_t[:]
        chunk_sub_idx: hsize_t[:]

        slices_cursor = 0
        for chunk_idxidx in range(n_sel_chunks):
            chunk_idx = self.chunk_indices[chunk_idxidx]
            chunk_start, chunk_stop = self._chunk_start_stop(chunk_idx)

            if self.is_ascending:
                # O(n), plus O(n ~ n*logn) for the searchsorted above
                start_idx = starts_stops_idxidx[chunk_idxidx, 0]
                stop_idx = starts_stops_idxidx[chunk_idxidx, 1]
                value_sub_idx = value_full_idx_v[start_idx:stop_idx]
                chunk_sub_idx = idx_v[start_idx:stop_idx]
            # TODO optimize monotonic descending
            else:
                # O(n^2)
                mask = (chunk_start <= self.idx) & (self.idx < chunk_stop)
                # Don't copy when converting from np.intp to uint64 on 64-bit platforms
                value_sub_idx = asarray(np.flatnonzero(mask), np_hsize_t)
                chunk_sub_idx = self.idx[value_sub_idx]

            if max_rows > n_sel_chunks:
                chunk_to_slices_v[chunk_idxidx] = slices_cursor

            next_slices_cursor = _fancy_idx_to_slices(
                value_sub_idx, chunk_sub_idx, slices_v, slices_cursor
            )
            for i in range(slices_cursor, next_slices_cursor):
                slices_v[i, 0] -= chunk_start
            slices_cursor = next_slices_cursor

        if max_rows > n_sel_chunks:
            if slices_cursor == n_sel_chunks:
                # e.g. idx=[0, 2, 3], chunk_size=2, dset_size=4
                chunks_to_slices = None
            else:
                chunk_to_slices_v[n_sel_chunks] = slices_cursor
            # This is a view that wastes some memory. No point calling realloc() as this
            # object is volatile and will be dereferenced at the end of the current
            # __getitem__ / __setitem__ call.
            slices = slices[:slices_cursor]
        else:
            assert slices_cursor == n_sel_chunks
        return slices, chunks_to_slices

    @cython.cfunc
    def _chunk_sizes_in_chunk_indices(self):  # -> NDArray[np_hsize_t] | int:
        """Return the number of points taken from each chunk within self.chunk_indices.
        All but the last chunk always contain self.chunk_size points.
        """
        indices = self.chunk_indices
        idx_len = len(indices)
        assert idx_len > 0  # Empty slices are filtered out by index_chunk_mappers()

        if indices[idx_len - 1] < self.dset_size // self.chunk_size:
            # Not the last chunk, or last chunk but exactly divisible by chunk_size
            return self.chunk_size

        out = np.full_like(indices, fill_value=self.chunk_size)
        out[idx_len - 1] = self.last_chunk_size
        return out

    @cython.ccall
    def chunks_indexer(self):
        indices = np.asarray(self.chunk_indices)
        return _maybe_array_idx_to_slice(indices)

    @cython.ccall
    def whole_chunks_idxidx(self):
        # Don't double count when the same index is picked twice
        idx_unique = np.unique(self.idx)
        _, counts = np.unique(idx_unique // self.chunk_size, return_counts=True)
        return _maybe_array_idx_to_slice(counts == self._chunk_sizes_in_chunk_indices())


@cython.cfunc
@cython.nogil
@cython.exceptval(check=False)
def _fancy_idx_to_slices(
    value_sub_idx: hsize_t[:],
    chunk_sub_idx: hsize_t[:],
    slices: hsize_t[:, :],
    cursor: hsize_t,
) -> hsize_t:
    """Break down a one-dimensional integer array into a (greedy) minimal list of
    slices. Populate input of read_many_slices().

    Parameters
    ----------
    value_sub_idx:
        integer array index selecting the __getitem__ output value or __setitem__
        parameter value for the chunk
    chunk_sub_idx:
        integer array index selecting the points within the chunk
    slices:
        Return value of IndexChunkMapper._init_many_slices_params()
    cursor:
        Current row of slices

    Returns
    -------
    Next row of slices
    """
    npoints = len(value_sub_idx)

    i = 0
    while i < npoints:
        value_sub_start = value_sub_idx[i]
        chunk_sub_start = chunk_sub_idx[i]
        count = 1
        value_sub_stride = 1  # This is 1 for monotonically increasing indices
        chunk_sub_stride = 1

        while True:
            if i == npoints - 1:
                break

            v0 = value_sub_idx[i]
            v1 = value_sub_idx[i + 1]
            c0 = chunk_sub_idx[i]
            c1 = chunk_sub_idx[i + 1]

            if v1 <= v0 or c1 <= c0:
                break  # step < 1

            if count == 1:
                # Recognize a stride pattern
                value_sub_stride = v1 - v0
                chunk_sub_stride = c1 - c0

                # Handle special case where having stride>1 now would break
                # a contiguous block with stride=1 immediately afterwards
                if i < npoints - 2:
                    v2 = value_sub_idx[i + 2]
                    if value_sub_stride > 1 and v2 == v1 + 1:
                        value_sub_stride = 1
                        chunk_sub_stride = 1
                        break

                    c2 = chunk_sub_idx[i + 2]
                    if chunk_sub_stride > 1 and c2 == c1 + 1:
                        value_sub_stride = 1
                        chunk_sub_stride = 1
                        break

            else:
                if v1 - v0 != value_sub_stride:
                    break
                if c1 - c0 != chunk_sub_stride:
                    break

            count += 1
            i += 1
        i += 1

        # For reference, see subchunk_map.pxd::ReadManySlicesColumn
        slices[cursor, 0] = chunk_sub_start
        slices[cursor, 1] = value_sub_start
        slices[cursor, 2] = count
        slices[cursor, 3] = chunk_sub_stride
        slices[cursor, 4] = value_sub_stride
        cursor += 1

    return cursor


def index_chunk_mappers(
    idx: Any,
    shape: tuple[int, ...],
    chunk_size: tuple[int, ...] | ChunkSize,
) -> tuple[Tuple, list[IndexChunkMapper]]:
    """Preprocess a numpy fancy index used in __getitem__ or __setitem__

    Returns
    -------
    - ndindex.Tuple with the preprocessed index
    - list of IndexChunkMapper objects, one per axis
      (including those omitted in the index)
    """
    assert isinstance(chunk_size, (tuple, ChunkSize))
    if not all(c > 0 for c in chunk_size):
        raise ValueError("chunk sizes must be structly positive")

    if isinstance(idx, Tuple):
        pass
    elif isinstance(idx, tuple):
        idx = Tuple(*idx)
    else:
        idx = Tuple(ndindex(idx))

    assert isinstance(shape, tuple)
    if any(dim < 0 for dim in shape):
        raise ValueError("shape dimensions must be non-negative")
    if len(shape) != len(chunk_size):
        raise ValueError("chunks dimensions must equal the array dimensions")

    if idx.isempty(shape):
        # abort early for empty index
        return idx, []

    idx_len = len(idx.args)

    d: hsize_t
    n: hsize_t
    fancy_count = 0
    mappers = []

    # Process the prefix of the axes which idx selects on
    for i, d, n in zip(idx.args, shape[:idx_len], chunk_size[:idx_len]):
        if isinstance(i, ellipsis):  # ndindex wrapper for Ellipsis; reduces erratically
            raise NotImplementedError("Ellipsis not supported")

        i = i.reduce((d,)).raw

        # _index_to_mapper tentatively simplifies fancy indices to slices.
        # However, it would be a mistake to simplify [[0, 1], [0, 1]] to [:2, :2]!
        if isinstance(i, np.ndarray):
            fancy_count += 1
            if fancy_count > 1:
                raise NotImplementedError("Multiple fancy indices")

        mapper = _index_to_mapper(i, d, n)
        mappers.append(mapper)

    # Handle the remaining suffix axes on which we did not select, we still need to
    # break them up into chunks.
    for d, n in zip(shape[idx_len:], chunk_size[idx_len:]):
        mappers.append(SliceMapper(slice(0, d, 1), d, n))

    return idx, mappers


@cython.cfunc
def _index_to_mapper(idx, dset_size: hsize_t, chunk_size: hsize_t) -> IndexChunkMapper:
    """Convert a one-dimensional index, preprocessed by ndindex, to a mapper"""
    if isinstance(idx, int):
        return IntegerMapper(idx, dset_size, chunk_size)

    if isinstance(idx, np.ndarray):
        if idx.ndim != 1:
            raise NotImplementedError("array index must be 1-dimensional")
        idx = _maybe_array_idx_to_slice(idx)
        if isinstance(idx, np.ndarray):
            return IntegerArrayMapper(idx, dset_size, chunk_size)

    if isinstance(idx, slice):
        return SliceMapper(idx, dset_size, chunk_size)

    if idx is None:
        raise NotImplementedError(f"None/np.newaxis not supported")

    # Unreachable: ndindex preprocessing should have caught this
    raise TypeError(f"Bad index {idx} of type {type(idx)}")  # pragma: nocover


@cython.cfunc
def _maybe_array_idx_to_slice(idx: Any):  # -> NDArray[np_hsize_t] | slice:
    """Attempt to convert an integer or boolean array index to a slice"""
    # boolean array indices can be trivially expressed as an integer array
    if idx.dtype == bool:
        idx = np.flatnonzero(idx)

    # Don't copy when converting from np.intp to uint64 on 64-bit platforms
    idx = asarray(idx, np_hsize_t)

    if not idx.flags.writeable:
        # Cython doesn't support read-only views in pure Python mode,
        # so we're forced to copy to prevent a crash in the next line.
        idx = idx.copy()

    idx_v: hsize_t[:] = idx
    idx_len = len(idx_v)
    if idx_len == 0:
        return slice(0, 0, 1)

    start = idx_v[0]
    if idx_len == 1:
        return slice(int(start), int(start + 1), 1)
    if idx_v[1] <= start:  # step <1
        return idx
    stop = idx_v[idx_len - 1] + 1
    step = idx_v[1] - start
    if idx_len == 2:
        return slice(int(start), int(stop), int(step))

    if stop2count(start, stop, step) == idx_len:
        j = start + step + step
        for i in range(2, idx_len):
            if idx_v[i] != j:
                break
            j += step
        else:
            return slice(int(start), int(stop), int(step))

    return idx


class TransferType(enum.Enum):
    getitem = 0
    setitem = 1
    slab_to_slab = 2


@cython.ccall
def read_many_slices_params_nd(
    transfer_type: TransferType,
    mappers: list[IndexChunkMapper],
    chunk_idxidx: hsize_t[:, :],
    src_slab_offsets: hsize_t[:],
    dst_slab_offsets: hsize_t[:],
) -> hsize_t[:, :, :]:
    """Assemble the parameters for read_many_slices() for a multi-dimensional index
    for a specific transfer type between two arrays.

    Parameters
    ----------
    transfer_type:
        The type of transfer to perform:

        TransferType.getitem
            The src array of read_many_slices() is an individual slab
            The dst array is the return value of __getitem__
            src_slab_offsets must have as many rows as chunk_idxidx
        TransferType.setitem
            The src array is the value parameter of __setitem__
            The dst array is an individual slab
            dst_slab_offsets must have as many rows as chunk_idxidx
        TransferType.slab_to_slab
            Both src and dst arrays are slabs

    mappers:
        List of IndexChunkMapper objects, one per axis of the index, as returned by
        index_chunk_mappers()
    chunk_idxidx:
        2D array of shape (nchunks, ndim) where each row represents a chunk to be
        transferred and point contains the index of IndexChunkMapper.chunk_indices for
        the axis matching the column. In other words,

            mappers[j].chunk_indices[chunk_idxidx[i, j]] * mappers[j].chunk_size

        is the address along axis j of the top-left corner in the virtual array
        of the i-th chunk to be transferred.

    src_slab_offsets:
        1D array of shape (nchunks, ) with the offset of the slab along axis 0, to be
        added to addresses relative to the start of the chunk in the src slab.
        Ignored when transfer_type==TransferType.setitem.
    dst_slab_offsets:
        1D array of shape (nchunks, ) with the offset of the slab along axis 0, to be
        added to addresses relative to the start of the chunk in the dst slab.
        Ignored when transfer_type==TransferType.getitem.

    Returns
    -------
    3D array of indices of shape (nslices, 5, ndim), ready to be passed to
    read_many_slices().

    There is one row per slice.
    In absence of fancy indexing, this is the same as the number of chunks.

    Axis 1 contains exactly 5 columns:

    - src_start
    - dst_start
    - count
    - src_stride
    - dst_stride

    See slicetools.pyx::read_many_slices() for documentation on these parameters.
    See subchunk_map.pxd::ReadManySlicesNDColumn for reference on the column indices.

    The final column represents the axis. Note how axis must be last, and thus
    internally C-contiguous, in order to allow read_many_slices() to use the index as-is
    without any further processing.
    """
    # Hack around lack of cython enums in pure python mode
    transfer_type_int: cython.int = transfer_type.value
    TRANSFER_TYPE_GETITEM: cython.int = TransferType.getitem.value
    TRANSFER_TYPE_SETITEM: cython.int = TransferType.setitem.value

    nchunks = chunk_idxidx.shape[0]
    ndim = chunk_idxidx.shape[1]

    # Store the slices in a C array (in the stack) of views (in the heap).
    # In order to keep the reference counter of the views, which are Python objects,
    # above zero for the duration of the function, we also need to reference them from a
    # Python object.
    slices_refs = []
    # Can't define module-level constants in pure-python mode
    # https://github.com/numpy/numpy/blob/b0a52b76c2bebe5c237722589d13bf02affa9c43/numpy/core/include/numpy/ndarraytypes.h#L30
    # NPY_MAXDIMS = 32
    slices: hsize_t[:, :][32]  # type: ignore
    if not cython.compiled:
        slices = [None] * ndim

    # Generate the slices along each axis and calculate the flattened number of slices
    # to return. This is trivial in absence of a fancy index as there is a 1:1
    # relationship between chunks and slices.
    nslices: ssize_t = nchunks
    fancy_axis = -1
    for j in range(ndim):
        mapper: IndexChunkMapper = mappers[j]
        slices_j: hsize_t[:, :]
        slices_j, c2s = mapper.read_many_slices_params()
        slices[j] = slices_j
        slices_refs.append(slices_j)

        assert slices_j.shape[1] == 5
        if c2s is None:
            # Basic index, or fancy index that could be coerced into 1 slice per chunk
            assert slices_j.shape[0] == len(mapper.chunk_indices)
        else:
            # There is a fancy index on this axis. Note that fancy indices can return
            # c2s=None and effectively declass themselves to basic indices for the
            # purpose of this algorithm when they can be represented as exactly one
            # slice per chunk; e.g. idx=[0, 2, 3], chunk_size=2, dset_size=4.

            assert fancy_axis == -1  # There can be at most one fancy index
            fancy_axis = j
            chunks_to_slices: hsize_t[:] = c2s
            assert len(chunks_to_slices) == len(mapper.chunk_indices) + 1

            # Recalculate nslices
            nslices = 0
            for i in range(nchunks):
                k = chunk_idxidx[i, j]
                if not cython.compiled:
                    k = int(k)  # uint64_t + int -> float64
                fancy_count = chunks_to_slices[k + 1] - chunks_to_slices[k]
                if not cython.compiled:
                    fancy_count = int(fancy_count)  # uint64_t + int -> float64
                nslices += fancy_count

    # Allocate memory for return value
    ndslices: hsize_t[:, :, :] = np.empty((nslices, 5, ndim), dtype=np_hsize_t)

    # Perform pseudo-cartesian product of mapper.read_many_slices_params()[0]
    # for each axis. Unlike a true cartesian product, chunk_idxidx can skip chunks.
    #
    # A helpful legend to this seven-counters monstruosity,
    # with many apologies to the reader:
    #
    # i: row of chunks; the flat i-th chunk caught not only by the intersection of
    #    mappers[*].chunk_indices, but also by the per-chunk mask on slab_indices.
    # j: axis; column of chunks
    # k: the index on mappers[j].chunk_indices.
    #    mappers[j].chunk_indices[k] * chunk_size[j] is the address of the
    #    top-left corner of the chunk along axis j in the virtual dataset
    #
    # Additional counters only for when there's a fancy index, which causes a 1:N
    # relationship between chunks and slices:
    # fancy_k:   like k, but always pinned to the fancy index regardless of the
    #            current axis
    # fancy_cur: the current slice along the fancy index, within the current chunk k
    # outs:      the n-th n-dimensional slice to process with read_many_slices;
    #            cursor on axis 0 of the return value
    # inps:      the n-th 1-dimensional slice on slices[j].
    #            When j == fancy_axis, this equals to fans; on all other axes, this is a
    #            repetition of k for as many slices there are in the fancy index for the
    #            current chunk.
    with cython.nogil:
        for j in range(ndim):
            slices_j = slices[j]
            # For the mapping, refer to docstring of
            # IndexChunkMapper.read_many_slices_params()
            # For column indices, refer to subchunk_map.pxd::ReadManySlicesColumn
            if transfer_type_int == TRANSFER_TYPE_GETITEM:
                src_start = slices_j[:, 0]  # chunk_sub_start
                dst_start = slices_j[:, 1]  # value_sub_start
                count = slices_j[:, 2]  # count
                src_stride = slices_j[:, 3]  # chunk_sub_stride
                dst_stride = slices_j[:, 4]  # value_sub_stride
                add_src_offset: bint = j == 0
                add_dst_offset: bint = False

            elif transfer_type_int == TRANSFER_TYPE_SETITEM:
                src_start = slices_j[:, 1]  # value_sub_start
                dst_start = slices_j[:, 0]  # chunk_sub_start
                count = slices_j[:, 2]  # count
                src_stride = slices_j[:, 4]  # value_sub_stride
                dst_stride = slices_j[:, 3]  # chunk_sub_stride
                add_src_offset = False
                add_dst_offset = j == 0

            else:  # slab-to-slab
                src_start = slices_j[:, 0]  # chunk_sub_start
                dst_start = slices_j[:, 0]  # chunk_sub_start
                count = slices_j[:, 2]  # count
                src_stride = slices_j[:, 3]  # chunk_sub_stride
                dst_stride = slices_j[:, 3]  # chunk_sub_stride
                add_src_offset = j == 0
                add_dst_offset = j == 0

            assert not add_src_offset or len(src_slab_offsets) == nchunks
            assert not add_dst_offset or len(dst_slab_offsets) == nchunks

            if fancy_axis == -1:
                # 1 chunk : 1 slice
                for i in range(nchunks):
                    k = chunk_idxidx[i, j]
                    src_offset = src_slab_offsets[i] if add_src_offset else 0
                    dst_offset = dst_slab_offsets[i] if add_dst_offset else 0

                    # For reference, see subchunk_map.pxd::ReadManySlicesNDColumn
                    ndslices[i, 0, j] = src_start[k] + src_offset
                    ndslices[i, 1, j] = dst_start[k] + dst_offset
                    ndslices[i, 2, j] = count[k]
                    ndslices[i, 3, j] = src_stride[k]
                    ndslices[i, 4, j] = dst_stride[k]

            else:
                # 1 chunk : n slices for the fancy index
                # 1 chunk : 1 slice on other axes, which must be replicated
                # to match the fancy index
                outs: ssize_t = 0
                for i in range(nchunks):
                    if j != fancy_axis:
                        k = chunk_idxidx[i, j]
                    src_offset = src_slab_offsets[i] if add_src_offset else 0
                    dst_offset = dst_slab_offsets[i] if add_dst_offset else 0

                    fancy_k = chunk_idxidx[i, fancy_axis]
                    if not cython.compiled:
                        fancy_k = int(fancy_k)  # uint64_t + int -> float64
                    fancy_start = chunks_to_slices[fancy_k]
                    fancy_stop = chunks_to_slices[fancy_k + 1]
                    if not cython.compiled:
                        fancy_start = int(fancy_start)
                        fancy_stop = int(fancy_stop)

                    for fancy_cur in range(fancy_start, fancy_stop):
                        inps = fancy_cur if j == fancy_axis else k
                        # For reference, see subchunk_map.pxd::ReadManySlicesNDColumn
                        ndslices[outs, 0, j] = src_start[inps] + src_offset
                        ndslices[outs, 1, j] = dst_start[inps] + dst_offset
                        ndslices[outs, 2, j] = count[inps]
                        ndslices[outs, 3, j] = src_stride[inps]
                        ndslices[outs, 4, j] = dst_stride[inps]
                        outs += 1

                assert outs == nslices

    return ndslices


def as_subchunk_map(
    idx: Any,
    shape: tuple[int, ...],
    chunk_size: tuple[int, ...] | ChunkSize,
) -> Iterator[
    tuple[
        Tuple,
        AnySlicerND,
        AnySlicerND,
    ]
]:
    """Computes the chunk selection assignment. In particular, given a `chunk_size`
    it returns triple (chunk_idx, value_sub_idx, chunk_sub_idx) such that for a
    chunked Dataset `ds` we can translate selections like

    >> value = ds[idx]

    into selecting from the individual chunks of `ds` as

    >> value = np.empty(output_shape)
    >> for chunk_idx, value_sub_idx, chunk_sub_idx in as_subchunk_map(
    ..     idx, ds.shape, ds.chunk_size
    .. ):
    ..     value[value_sub_idx] = ds.data_dict[chunk_idx][chunk_sub_idx]

    Similarly, assignments like

    >> ds[idx] = value

    can be translated into

    >> for chunk_idx, value_sub_idx, chunk_sub_idx in as_subchunk_map(
    ..     idx, ds.shape, ds.chunk_size
    .. ):
    ..     ds.data_dict[chunk_idx][chunk_sub_idx] = value[value_sub_idx]

    :param idx: the "index" to read from / write to the Dataset
    :param shape: the shape of the Dataset
    :param chunk_size: the `ChunkSize` of the Dataset
    :return: a generator of `(chunk_idx, value_sub_idx, chunk_sub_idx)` tuples
    """
    idx, mappers = index_chunk_mappers(idx, shape, chunk_size)
    if not mappers:
        return
    idx_len = len(idx.args)

    mapper: IndexChunkMapper  # noqa: F842
    chunk_subindexes = [
        [mapper.chunk_submap(chunk_idx) for chunk_idx in mapper.chunk_indices]
        for mapper in mappers
    ]

    # Now combine the chunk_slices and subindexes for each dimension into tuples
    # across all dimensions.
    for p in itertools.product(*chunk_subindexes):
        chunk_slices, value_subidxs, chunk_subidxs = zip(*p)

        # skip dimensions which were sliced away
        value_subidxs = tuple(
            value_subidx
            for value_subidx in value_subidxs
            if value_subidx is not DROP_AXIS
        )
        # skip suffix dimensions
        chunk_subidxs = chunk_subidxs[:idx_len]

        yield Tuple(*chunk_slices), value_subidxs, chunk_subidxs



from typing import Any

import hypothesis
import numpy as np
import pytest
from h5py._hl.selections import Selection
from hypothesis import given
from hypothesis import strategies as st
from numpy.testing import assert_equal

from .. import VersionedHDF5File
from ..cytools import count2stop
from ..slicetools import (
    build_slab_indices_and_offsets,
    read_many_slices,
    spaceid_to_slice,
)

max_examples = 10_000


def test_spaceid_to_slice(h5file):
    shape = 10
    a = h5file.create_dataset("a", data=np.arange(shape))

    for start in range(0, shape):
        for count in range(0, shape):
            for stride in range(1, shape):
                for block in range(0, shape):
                    if count != 1 and block != 1:
                        # Not yet supported. Doesn't seem to be supported
                        # by HDF5 either (?)
                        continue

                    spaceid = a.id.get_space()
                    spaceid.select_hyperslab((start,), (count,), (stride,), (block,))
                    sel = Selection((shape,), spaceid)
                    try:
                        a[sel]
                    except (ValueError, OSError):
                        # HDF5 doesn't allow stride/count combinations
                        # that are impossible (the count must be the exact
                        # number of elements in the selected block).
                        # Rather than trying to enumerate those here, we
                        # just check what doesn't give an error.
                        continue
                    try:
                        s = spaceid_to_slice(spaceid)
                    except:
                        print(start, count, stride, block)
                        raise
                    assert_equal(a[s.raw], a[sel], f"{(start, count, stride, block)}")


def test_build_slab_indices_and_offsets_dense(h5file):
    chunks = (2, 3)
    data = np.array(
        [
            [1, 2, 3, 4, 5],
            [6, 7, 8, 9, 10],
            [11, 12, 13, 14, 15],
        ]
    )
    expect_raw_data = np.array(
        [
            # chunk (0, 0), offset 0
            [1, 2, 3],
            [6, 7, 8],
            # chunk (0, 1), offset 2
            [4, 5, 0],
            [9, 10, 0],
            # chunk (1, 0), offset 4
            [11, 12, 13],
            [0, 0, 0],
            # chunk (1, 1), offset 6
            [14, 15, 0],
            [0, 0, 0],
        ]
    )
    vf = VersionedHDF5File(h5file)
    with vf.stage_version("r0") as sv:
        sv.create_dataset("a", data=data, chunks=chunks)
    virt_dset = h5file["_version_data/versions/r0/a"]
    raw_data = h5file["_version_data/a/raw_data"]
    np.testing.assert_array_equal(virt_dset[:], data, strict=True)
    np.testing.assert_array_equal(raw_data[:], expect_raw_data, strict=True)

    dcpl = virt_dset.id.get_create_plist()
    indices, offsets = build_slab_indices_and_offsets(dcpl, data.shape, chunks)
    np.testing.assert_array_equal(indices, [[1, 1], [1, 1]])
    np.testing.assert_array_equal(offsets, [[0, 2], [4, 6]])


def test_build_slab_indices_and_offsets_sparse(h5file):
    chunks = (2, 3)
    shape = (3, 5)
    vf = VersionedHDF5File(h5file)
    with vf.stage_version("r0") as sv:
        sv.create_dataset(
            "a", shape=shape, data=None, chunks=chunks, dtype=np.int64, fillvalue=123
        )
    virt_dset = h5file["_version_data/versions/r0/a"]
    expect = np.full(shape, fill_value=123, dtype=np.int64)
    np.testing.assert_array_equal(
        virt_dset[:], np.full(shape, fill_value=123, dtype=np.int64), strict=True
    )

    dcpl = virt_dset.id.get_create_plist()
    indices, offsets = build_slab_indices_and_offsets(dcpl, shape, chunks)
    np.testing.assert_array_equal(indices, [[0, 0], [0, 0]])
    np.testing.assert_array_equal(offsets, [[0, 0], [0, 0]])

    # This test is recursive, as in order to create r1 on disk versioned_hdf5 is going
    # to call build_slab_indices_and_offsets.
    with vf.stage_version("r1") as sv:
        dset = sv["a"]
        dset[1, 3] = 456
        dset[2, 0] = 789

    virt_dset = h5file["_version_data/versions/r1/a"]
    raw_data = h5file["_version_data/a/raw_data"]

    expect[1, 3] = 456
    expect[2, 0] = 789
    np.testing.assert_array_equal(virt_dset[:], expect, strict=True)
    expect_raw = np.array(
        [
            [123, 123, 123],
            [456, 123, 123],
            [789, 123, 123],
            [123, 123, 123],
            # FIXME Spurious extra chunk. Discussion:
            # https://github.com/deshaw/versioned-hdf5/pull/385#discussion_r1817313138
            [123, 123, 123],
            [123, 123, 123],
        ],
        dtype=np.int64,
    )
    np.testing.assert_array_equal(raw_data[:], expect_raw, strict=True)

    dcpl = virt_dset.id.get_create_plist()
    indices, offsets = build_slab_indices_and_offsets(dcpl, shape, chunks)
    np.testing.assert_array_equal(indices, [[0, 1], [1, 0]])
    np.testing.assert_array_equal(offsets, [[0, 0], [2, 0]])


@st.composite
def bound_slices_st(draw, arr_size: int, view_size: int) -> slice:
    """Hypothesis draw of a slice object to slice an array of <arr_size> points
    along an axis, returning a view of <view_size> points::

        arr = np.empty(arr_size)
        idx = bound_slices_st(arr_size, view_size).example()
        assert arr[idx].size == view_size
    """
    start = draw(st.integers(0, arr_size - view_size))

    # Don't want to figure out the exact formula to avoid off-by-one errors.
    # Just add some margin and then count down.
    for max_step in range((arr_size - start + 1) // view_size + 1, 0, -1):
        tmp_stop = count2stop(start, view_size, max_step)
        if tmp_stop <= arr_size:
            break

    step = draw(st.integers(1, max_step))
    stop = count2stop(start, view_size, step)
    assert len(range(start, min(stop, arr_size), step)) == view_size
    return slice(start, stop, step)


@st.composite
def matching_slices_st(draw, src_size: int, dst_size: int) -> tuple[slice, slice]:
    """Hypothesis draw to slice src and dst arrays along the same axis
    dst[idx_dst] = src[idx_src]

    Returns tuples of idx_src, idx_dst
    where src[idx_src].size == dst[idx_dst].size
    """
    view_size = draw(st.integers(1, min(src_size, dst_size)))
    slices_st = st.tuples(
        bound_slices_st(src_size, view_size),
        bound_slices_st(dst_size, view_size),
    )
    return draw(slices_st)


@st.composite
def many_slices_st(
    draw, max_ndim: int = 4, max_size: int = 20, max_nslices: int = 4
) -> tuple[
    # shape of src array
    tuple[int, ...],
    # shape of dst array, with same dimensionality as src
    tuple[int, ...],
    # list of indices to slice src array with
    list[tuple[slice, ...]],
    # matching list of indices to slice dst array with
    list[tuple[slice, ...]],
]:
    src_shape_st = st.lists(st.integers(1, max_size), min_size=1, max_size=max_ndim)
    src_shape = tuple(draw(src_shape_st))
    ndim = len(src_shape)
    dst_shape_st = st.lists(st.integers(1, max_size), min_size=ndim, max_size=ndim)
    dst_shape = tuple(draw(dst_shape_st))

    nslices = draw(st.integers(1, max_nslices))
    src_indices = []
    dst_indices = []
    for _ in range(nslices):
        src_idx = []
        dst_idx = []
        for src_size, dst_size in zip(src_shape, dst_shape):
            src_slice, dst_slice = draw(matching_slices_st(src_size, dst_size))
            src_idx.append(src_slice)
            dst_idx.append(dst_slice)
        src_indices.append(tuple(src_idx))
        dst_indices.append(tuple(dst_idx))

    return src_shape, dst_shape, src_indices, dst_indices


@given(args=many_slices_st())
@hypothesis.settings(
    max_examples=max_examples,
    deadline=None,
    # h5file is not reset between hypothesis examples
    suppress_health_check=[hypothesis.HealthCheck.function_scoped_fixture],
)
def test_read_many_slices(h5file, args):
    shape_from, shape_to, indices_from, indices_to = args

    src = np.arange(1, np.prod(shape_from) + 1, dtype=np.int32).reshape(shape_from)

    expect = np.zeros(shape_to, dtype=src.dtype)
    src_start = []
    dst_start = []
    count = []
    src_step = []
    dst_step = []
    for idx_from, idx_to in zip(indices_from, indices_to):
        expect[idx_to] = src[idx_from]

        src_start.append([s.start for s in idx_from])
        src_step.append([s.step for s in idx_from])
        dst_start.append([s.start for s in idx_to])
        dst_step.append([s.step for s in idx_to])
        count.append([len(range(s.start, s.stop, s.step)) for s in idx_from])

    # Test numpy->numpy
    dst = np.zeros(shape_to, dtype=src.dtype)
    read_many_slices(src, dst, src_start, dst_start, count, src_step, dst_step)
    np.testing.assert_array_equal(dst, expect, strict=True)

    # h5file fixture is not reset between hypothesis examples
    if "a" in h5file:
        del h5file["a"]
    dset = h5file.create_dataset("a", data=src)

    # Test h5py->numpy
    for kwargs in ({}, {"fast": True}, {"fast": False}):
        dst = np.zeros(shape_to, dtype=src.dtype)
        read_many_slices(
            dset, dst, src_start, dst_start, count, src_step, dst_step, **kwargs
        )
        np.testing.assert_array_equal(dst, expect, strict=True)


@pytest.mark.parametrize("step", [1, 2, 3, 4, 5, 6])
@pytest.mark.parametrize("start", [0, 1, 2, 3, 4, 5, 6])
@pytest.mark.parametrize("use_h5", [True, False])
def test_read_many_slices_clip_src_count(h5file, use_h5, start, step):
    """Counts are clipped automatically because either
    src_start or src_stride are too large.
    This causes some slices to become size 0 along one dimension.
    """
    src = np.arange(1, 21).reshape(5, 4)
    src_view = src[start::step]
    expect = np.zeros((8, 4), dtype=src.dtype)
    expect[: len(src_view)] = src_view

    # Test that completely skipping a slice because its clipped count is zero doesn't
    # cause the next rows to be skipped
    expect[7, 3] = src[4, 3]

    if use_h5:
        src = h5file.create_dataset("a", data=src)

    dst = np.zeros_like(expect)
    read_many_slices(
        src,
        dst,
        src_start=[(start, 0), (4, 3)],
        dst_start=[(0, 0), (7, 3)],
        count=[(999, 999), (999, 999)],
        src_stride=[(step, 1), (1, 1)],
    )
    np.testing.assert_array_equal(dst, expect, strict=True)


@pytest.mark.parametrize("step", [1, 2, 3, 4, 5, 6])
@pytest.mark.parametrize("start", [0, 1, 2, 3, 4, 5, 6])
@pytest.mark.parametrize("use_h5", [True, False])
def test_read_many_slices_clip_dst_count(h5file, use_h5, start, step):
    """Counts are clipped automatically because either
    dst_start or dst_stride are too large.
    This causes some slices to become size 0 along one dimension.
    """
    src = np.arange(1, 41).reshape(10, 4)
    expect = np.zeros((5, 4), dtype=src.dtype)
    dst_view = expect[start::step]
    expect[start::step] = src[: len(dst_view)]

    # Test that completely skipping a slice because its clipped count is zero doesn't
    # cause the next slices to be skipped
    expect[4, 3] = src[7, 3]

    if use_h5:
        src = h5file.create_dataset("a", data=src)

    dst = np.zeros_like(expect)
    read_many_slices(
        src,
        dst,
        src_start=[(0, 0), (7, 3)],
        dst_start=[(start, 0), (4, 3)],
        count=[(999, 999), (999, 999)],
        src_stride=None,
        dst_stride=[(step, 1), (1, 1)],
    )
    np.testing.assert_array_equal(dst, expect, strict=True)


@pytest.mark.parametrize("all_1d", [True, False])
@pytest.mark.parametrize("use_h5", [True, False])
def test_read_many_slices_broadcast_indices(h5file, use_h5, all_1d):
    """1D indices are automatically broadcast to 2D"""
    src = np.arange(1, 41).reshape(10, 4)
    expect = np.zeros((5, 4), dtype=src.dtype)
    expect[0:3, 2:4] = src[1:4, 1:3]
    if not all_1d:
        expect[2:5, 1:3] = src[1:4, 1:3]

    if use_h5:
        src = h5file.create_dataset("a", data=src)

    dst = np.zeros_like(expect)
    read_many_slices(
        src,
        dst,
        src_start=(1, 1),
        dst_start=(0, 2) if all_1d else [(0, 2), (2, 1)],
        count=(3, 2),
    )
    np.testing.assert_array_equal(dst, expect, strict=True)


@pytest.mark.parametrize("use_h5", [True, False])
def test_read_many_slices_no_indices(h5file, use_h5):
    """Edge case where there are no slices of data to copy"""
    src = np.arange(1, 41).reshape(10, 4)
    dst = np.zeros((5, 4), dtype=src.dtype)
    if use_h5:
        src = h5file.create_dataset("a", data=src)

    read_many_slices(src, dst, [], [], [])
    assert (dst == 0).all()


@pytest.mark.parametrize("contiguous_cols", [True, False])
@pytest.mark.parametrize("use_h5", [True, False])
def test_read_many_slices_noncontiguous_idx(h5file, use_h5, contiguous_cols):
    """one or more coordinate array is not C-contiguous along the rows or both
    columns and rows.

    Non-contiguous rows are supported thanks to Cython strided views.
    Non-contiguous columns are not supported in hdf5 fast mode and are transparently
    deep-copied before use.
    """
    src = np.arange(1, 51).reshape(10, 5)

    src_start = np.array([(0, 1, 2), (9, 9, 9), (2, 3, 4)])
    if contiguous_cols:
        src_start = src_start[::2, :2]
        expect = [[src[0, 1]], [src[2, 3]]]
    else:
        src_start = src_start[::2, ::2]
        expect = [[src[0, 2]], [src[2, 4]]]

    assert not src_start.flags.c_contiguous

    if use_h5:
        src = h5file.create_dataset("a", data=src)

    dst = np.zeros((2, 1), dtype=src.dtype)
    read_many_slices(src, dst, src_start, [(0, 0), (1, 0)], (1, 1))
    np.testing.assert_equal(dst, expect)


@pytest.mark.parametrize("use_h5", [True, False])
def test_read_many_slices_src_size_zero(h5file, use_h5):
    """src array has size 0"""
    src = np.empty((3, 0))
    dst = np.zeros((3, 3), dtype=src.dtype)
    if use_h5:
        src = h5file.create_dataset("a", data=src)
    read_many_slices(src, dst, [(0, 0)], [(0, 0)], [(3, 3)])
    assert (dst == 0).all()


@pytest.mark.parametrize("use_h5", [True, False])
def test_read_many_slices_dst_size_zero(h5file, use_h5):
    """dst array has size 0"""
    src = np.arange(1, 10).reshape(3, 3)
    dst = np.empty((0, 3), dtype=src.dtype)
    if use_h5:
        src = h5file.create_dataset("a", data=src)
    read_many_slices(src, dst, [(0, 0)], [(0, 0)], [(3, 3)])


def test_read_many_slices_noncontiguous_dst(h5file):
    """dst is a numpy array that's not C-contiguous.
    It is supported, but only with fast=False.
    """
    src = np.arange(1, 17).reshape(4, 4)
    expect = np.zeros((8, 12), dtype=src.dtype)
    expect[::2, ::3] = src[()]

    dst = np.zeros_like(expect)
    read_many_slices(src, dst[::2, ::3], [(0, 0), (2, 0)], [(0, 0), (2, 0)], (2, 4))
    np.testing.assert_equal(dst, expect)

    src = h5file.create_dataset("a", data=src)
    dst = np.zeros_like(expect)
    read_many_slices(
        src, dst[::2, ::3], [(0, 0), (2, 0)], [(0, 0), (2, 0)], (2, 4), fast=False
    )
    np.testing.assert_equal(dst, expect)

    # h5py src only works in slow mode
    with pytest.raises(NotImplementedError, match="contiguous"):
        read_many_slices(
            src, dst[::2, ::3], [(0, 0), (2, 0)], [(0, 0), (2, 0)], (2, 4), fast=True
        )

    # For the sake of awareness, user must explicitly pass fast=False
    with pytest.raises(NotImplementedError, match="contiguous"):
        read_many_slices(src, dst[::2, ::3], [(0, 0), (2, 0)], [(0, 0), (2, 0)], (2, 4))


def test_read_many_slices_not_fast_read_ok(h5file):
    """src is a h5py dataset that doesn't support fast read"""
    src = np.array(["foo", "bar", "baz"]).astype(object)
    expect = np.array([b"bar"]).astype(object)
    src = h5file.create_dataset("a", data=src)
    assert not src._fast_read_ok

    dst = np.zeros_like(expect)
    read_many_slices(src, dst, [[1]], [[0]], [[1]])
    np.testing.assert_array_equal(dst, expect, strict=True)

    dst = np.zeros_like(expect)
    read_many_slices(src, dst, [[1]], [[0]], [[1]], fast=False)
    np.testing.assert_array_equal(dst, expect, strict=True)

    with pytest.raises(ValueError, match="fast transfer is not possible"):
        read_many_slices(src, dst, [[1]], [[0]], [[1]], fast=True)


def test_read_many_slices_fail():
    src = np.arange(1, 5)
    dst = np.zeros(4, dtype=src.dtype)

    # Negative indices in a list (np.asarray with unsigned dtype fails)
    with pytest.raises(OverflowError):
        read_many_slices(src, dst, [[-1]], [[0]], [[1]])
    # Negative indices in numpy array (needs explicit validation)
    with pytest.raises(OverflowError):
        read_many_slices(src, dst, np.array([[-1]]), [[0]], [[1]])

    # Stride 0
    with pytest.raises(ValueError, match="Strides must be strictly greater than zero"):
        read_many_slices(src, dst, [[0]], [[0]], [[1]], [[0]], [[1]])
    with pytest.raises(ValueError, match="Strides must be strictly greater than zero"):
        read_many_slices(src, dst, [[0]], [[0]], [[1]], [[1]], [[0]])

    # src/dst ndim=0
    with pytest.raises(ValueError, match="at least one dimension"):
        read_many_slices(np.array(0), np.array(0), [[]], [[]], [[]])

    # src.ndim != dst.ndim
    with pytest.raises(ValueError, match="same dtype and dimensionality"):
        read_many_slices(src.reshape(2, 2), dst, [[0]], [[0]], [[1]])

    # src.dtype != dst.dtype
    with pytest.raises(ValueError, match="same dtype and dimensionality"):
        read_many_slices(src.astype(float), dst, [[0]], [[0]], [[1]])

    # Fail to broadcast coordinates
    with pytest.raises(ValueError, match="cannot be broadcast"):
        read_many_slices(src, dst, [[0], [1]], [[0], [1], [2]], [[1]])

    # 0-sized coordinates
    with pytest.raises(ValueError, match="must have 1 or 2 dimensions"):
        read_many_slices(src, dst, 0, 0, 1)

    # coordinates with 3+ dims
    with pytest.raises(ValueError, match="must have 1 or 2 dimensions"):
        read_many_slices(src, dst, np.zeros((1, 1, 1)), [[0]], [[1]])

    # indices.shape[1] != src.ndim
    with pytest.raises(ValueError, match="as many columns as src.ndim"):
        read_many_slices(src, dst, [(0, 0)], [(0, 0)], [(1, 1)])

    # force fast transfer on numpy src
    with pytest.raises(ValueError, match="fast transfer is not possible"):
        read_many_slices(src, dst, [[0]], [[0]], [[1]], fast=True)

    # dst is not writeable
    dst.setflags(write=False)
    with pytest.raises(ValueError, match="writeable"):
        read_many_slices(src, dst, [[0]], [[0]], [[1]])



from __future__ import absolute_import, division, print_function, with_statement

import shutil
import tempfile
from contextlib import contextmanager


@contextmanager
def temp_dir_ctx():
    tmp_dir = tempfile.mkdtemp()
    yield tmp_dir
    shutil.rmtree(tmp_dir)



"""
Wrappers of h5py objects that work in memory

Much of this code is modified from code in h5py. See the LICENSE file for the
h5py license.
"""

from __future__ import annotations

import posixpath
import textwrap
import warnings
from collections import defaultdict
from typing import Iterable
from weakref import WeakValueDictionary

import numpy as np
from h5py import Dataset, Datatype, Empty, Group
from h5py import __version__ as h5py_version
from h5py import h5a, h5d, h5g, h5i, h5p, h5r, h5s, h5t
from h5py._hl import filters
from h5py._hl.base import guess_dtype, phil, with_phil
from h5py._hl.dataset import _LEGACY_GZIP_COMPRESSION_VALS
from h5py._hl.selections import guess_shape
from ndindex import (
    BooleanArray,
    ChunkSize,
    Integer,
    IntegerArray,
    Slice,
    Tuple,
    ndindex,
)

from .backend import DEFAULT_CHUNK_SIZE
from .slicetools import build_data_dict
from .subchunk_map import as_subchunk_map

_groups = WeakValueDictionary({})


class InMemoryGroup(Group):
    def __new__(cls, bind: h5g.GroupID, _committed: bool = False):
        # Make sure each group only corresponds to one InMemoryGroup instance.
        # Otherwise a new instance would lose track of any datasets or
        # subgroups created in the old one.
        if bind in _groups:
            return _groups[bind]
        obj = super().__new__(cls)
        obj._initialized = False
        _groups[bind] = obj
        return obj

    def __init__(self, bind: h5g.GroupID, _committed: bool = False):
        """Create a new InMemoryGroup object by binding to a low-level GroupID.

        Parameters
        ----------
        bind : h5g.GroupID
            Low-level GroupID to bind to
        _committed : bool
            True if the group has already been committed, False otherwise.
        """
        if self._initialized:
            return
        self._data = {}
        self._subgroups = {}
        self._chunks = defaultdict(type(None))
        self._compression = defaultdict(type(None))
        self._compression_opts = defaultdict(type(None))
        self._parent = None
        self._initialized = True
        self._committed = _committed
        super().__init__(bind)

    def close(self):
        """Mark self and any subgroups as committed."""
        self._committed = True
        for name in self:
            obj = self[name]
            if isinstance(obj, InMemoryGroup):
                obj.close()

    # Based on Group.__repr__
    def __repr__(self):
        namestr = f'"{self.name}"' if self.name is not None else "(anonymous)"
        if not self:
            return "<Closed InMemoryGroup>"
        if self._committed:
            return f"<Committed InMemoryGroup {namestr}>"

        text = [f"<InMemoryGroup {namestr} ({len(list(self))} members)>"]
        for item in self.values():
            text.append(textwrap.indent(repr(item), prefix="  "))
        return "\n".join(text)

    def _check_committed(self):
        if self._committed:
            namestr = ('"%s"' % self.name) if self.name is not None else "(anonymous)"
            raise ValueError("InMemoryGroup %s has already been committed" % namestr)

    def __getitem__(self, name):
        dirname, basename = posixpath.split(name)
        if dirname:
            return self.__getitem__(dirname)[basename]

        if name in self._data:
            return self._data[name]
        if name in self._subgroups:
            return self._subgroups[name]

        # If the name doesn't exist in self._data (the "in-memory"
        # part of InMemoryGroup), retrieve it from the actual underlying
        # h5py.Group, (i.e. the file itself).
        res = super().__getitem__(name)
        if isinstance(res, Group):
            self._subgroups[name] = self.__class__(res.id)
            return self._subgroups[name]
        elif isinstance(res, Dataset):
            self._add_to_data(name, res)
            return self._data[name]
        else:
            raise NotImplementedError(f"Cannot handle {type(res)!r}")

    def __setitem__(self, name, obj):
        self._check_committed()
        self._add_to_data(name, obj)

    def _add_to_data(self, name, obj):
        dirname, basename = posixpath.split(name)
        if dirname:
            if dirname not in self:
                self.create_group(dirname)
            self[dirname][basename] = obj
            return

        if isinstance(obj, Dataset):
            wrapped_dataset = self._data[name] = DatasetWrapper(
                InMemoryDataset(obj.id, parent=self)
            )
            self.set_compression(name, wrapped_dataset.dataset.id.raw_data.compression)
            self.set_compression_opts(
                name, wrapped_dataset.dataset.id.raw_data.compression_opts
            )
        elif isinstance(obj, Group):
            self._subgroups[name] = InMemoryGroup(obj.id)
        elif isinstance(obj, InMemoryGroup):
            self._subgroups[name] = obj
        elif isinstance(obj, DatasetLike):
            self._data[name] = obj
            self.set_compression(name, obj.compression)
            self.set_compression_opts(name, obj.compression_opts)
        else:
            self._data[name] = InMemoryArrayDataset(name, np.asarray(obj), parent=self)

    def __delitem__(self, name):
        self._check_committed()
        dirname, basename = posixpath.split(name)
        if dirname:
            if not basename:
                del self[dirname]
            else:
                del self[dirname][basename]
            return

        if name in self._data:
            del self._data[name]
        elif name in self._subgroups:
            for i in self[name]:
                del self[name][i]
            del self._subgroups[name]
            super().__delitem__(name)
        else:
            raise KeyError(f"{name!r} is not in {self}")

    @property
    def parent(self):
        if self._parent is None:
            return super().parent
        return self._parent

    @parent.setter
    def parent(self, p):
        self._parent = p

    def create_group(self, name, track_order=None):
        self._check_committed()
        if name.startswith("/"):
            raise ValueError(
                "Root level groups cannot be created inside of versioned groups"
            )
        group = type(self)(super().create_group(name, track_order=track_order).id)
        g = group
        n = name
        while n:
            dirname, basename = posixpath.split(n)
            if not dirname:
                parent = self
            else:
                parent = type(self)(g.parent.id)
            parent._subgroups[basename] = g
            g.parent = parent
            g = parent
            n = dirname
        return group

    def create_dataset(
        self,
        name,
        shape=None,
        dtype=None,
        data: np.ndarray | None = None,
        fillvalue=None,
        **kwds,
    ):
        self._check_committed()
        dirname, data_name = posixpath.split(name)
        if dirname and dirname not in self:
            self.create_group(dirname)
        if "maxshape" in kwds and any(i != None for i in kwds["maxshape"]):
            warnings.warn(
                "The maxshape parameter is currently ignored for versioned datasets."
            )
        data = _make_new_dset(
            data=data, shape=shape, dtype=dtype, fillvalue=fillvalue, **kwds
        )
        if shape is None:
            shape = data.shape
        if fillvalue is not None and isinstance(data, np.ndarray):
            data = InMemoryArrayDataset(name, data, parent=self, fillvalue=fillvalue)
        chunks = kwds.get("chunks")
        if data is None:
            data = InMemorySparseDataset(
                name,
                shape=shape,
                dtype=dtype,
                parent=self,
                fillvalue=fillvalue,
                chunks=chunks,
            )
        if chunks in [True, None]:
            if len(shape) == 1:
                chunks = (DEFAULT_CHUNK_SIZE,)
            else:
                raise NotImplementedError(
                    "chunks must be specified for multi-dimensional datasets"
                )
        if isinstance(chunks, int) and not isinstance(chunks, bool):
            chunks = (chunks,)
        if len(shape) != len(chunks):
            raise ValueError(
                f"Dimensions of chunks ({chunks}) must equal the dimensions of the shape ({shape})"
            )
        if len(shape) == 0:
            raise NotImplementedError("Scalar datasets are not implemented.")

        self.set_chunks(name, chunks)
        self.set_compression(name, kwds.get("compression"))
        self.set_compression_opts(name, kwds.get("compression_opts"))
        self[name] = data
        if dtype is not None:
            self[name]._dtype = dtype
        return self[name]

    def __iter__(self):
        names = list(self._data) + list(self._subgroups)
        for i in super().__iter__():
            if i in names:
                names.remove(i)
            yield i
        for i in names:
            yield i

    def __contains__(self, item):
        item = item + "/"
        root = self.versioned_root.name + "/"
        if item.startswith(root):
            item = item[len(root) :]
            if not item.rstrip("/"):
                return self == self.versioned_root
        item = item.rstrip("/")
        dirname, data_name = posixpath.split(item)
        if dirname not in ["", "/"]:
            return dirname in self and data_name in self[dirname]
        for i in self:
            if i == item:
                return True
        return False

    def datasets(self):
        res = self._data.copy()

        def _get(name, item):
            if name in res:
                return
            if isinstance(item, (Dataset, DatasetLike, np.ndarray)):
                res[name] = item

        self.visititems(_get)

        return res

    @property
    def versioned_root(self):
        p = self
        while p._parent is not None:
            p = p._parent
        return p

    @property
    def chunks(self):
        return self._chunks

    # TODO: Can we generalize this, set_compression, and set_compression_opts
    # into a single method? Descriptors?
    def set_chunks(self, item, value):
        full_name = item
        p = self
        while p._parent:
            p._chunks[full_name] = value
            _, basename = posixpath.split(p.name)
            full_name = basename + "/" + full_name
            p = p._parent
        self.versioned_root._chunks[full_name] = value

        dirname, basename = posixpath.split(item)
        while dirname:
            self[dirname]._chunks[basename] = value
            dirname, b = posixpath.split(dirname)
            basename = posixpath.join(b, basename)

    @property
    def compression(self):
        return self._compression

    def set_compression(self, item, value):
        full_name = item
        p = self
        while p._parent:
            p._compression[full_name] = value
            _, basename = posixpath.split(p.name)
            full_name = basename + "/" + full_name
            p = p._parent
        self.versioned_root._compression[full_name] = value

        dirname, basename = posixpath.split(item)
        while dirname:
            self[dirname]._compression[basename] = value
            dirname, b = posixpath.split(dirname)
            basename = posixpath.join(b, basename)

    @property
    def compression_opts(self):
        return self._compression_opts

    def set_compression_opts(self, item, value):
        full_name = item
        p = self
        while p._parent:
            p._compression_opts[full_name] = value
            _, basename = posixpath.split(p.name)
            full_name = basename + "/" + full_name
            p = p._parent
        self.versioned_root._compression_opts[full_name] = value

        dirname, basename = posixpath.split(item)
        while dirname:
            self[dirname]._compression_opts[basename] = value
            dirname, b = posixpath.split(dirname)
            basename = posixpath.join(b, basename)

    def visititems(self, func):
        self._visit("", func)

    def _visit(self, prefix, func):
        for name in self:
            func(posixpath.join(prefix, name), self[name])
            if isinstance(self[name], InMemoryGroup):
                self[name]._visit(posixpath.join(prefix, name), func)

    # TODO: override other relevant methods here


def _make_new_dset(
    shape: int | tuple[int, ...] | None = None,
    dtype: np.dtype | None = None,
    data: np.ndarray | None = None,
    chunks: tuple[int, ...] | None = None,
    compression: int | str | bool | None = None,
    shuffle: bool | None = None,
    fletcher32: bool | None = None,
    maxshape: tuple[int, ...] | None = None,
    compression_opts: int | tuple[int, ...] | None = None,
    fillvalue: int | str | float | None = None,
    scaleoffset: bool | int | None = None,
    track_times: bool | None = None,
    external: Iterable[tuple[str, int, int]] | None = None,
    track_order: bool | None = None,
    dcpl: h5p.PropDCID | None = None,
) -> np.ndarray:
    """Create a new low-level dataset identifier.

    Based on h5py._hl.dataset.make_new_dset(), except it doesn't actually create
    the dataset, it just canonicalizes the arguments. Additionally, this function
    allows datasets which are smaller than the data in any dimension to be
    instantiated, whereas the upstream h5py version does not.

    See the LICENSE file for the h5py license.

    Parameters
    ----------
    shuffle : bool | None
        Whether to call dcpl.set_shuffle() on the underlying PropDCID
    fletcher32 : bool | None
        Whether to call dcpl.set_fletcher32(). Note that scale/offset following
        fletcher32 in the filter chain will (almost?) always triggers a read
        error, as most scale/offset settings are lossy. Since fletcher32 must
        come first (see comment in h5py._hl.filters.fill_dcpl) combination of
        fletcher32 and scale/offset is prohibited.
    maxshape : tuple[int, ...] | None
        Max shape of the dataste
    compression_opts : int | tuple[int, ...] | None
        Compression options passed to h5py._hl.filters.fill_dcpl
    fillvalue : int | str | float | None
        Value used to fill the parts of chunks that extend beyond the dataset
    scaleoffset : bool | int | None
        This must be an integer when it is not None or False, except for integral
        data, for which scaleoffset == True is permissible (will use
        SO_INT_MINBITS_DEFAULT)
    track_times : bool |  None
        Argument to pass to dcpl.set_obj_track_times(track_times)
    shape : int | tuple[int, ...] | None
        Shape of the dataset to create
    dtype : np.dtype | None
        Dtype of the dataset to create
    data : np.ndarray | None
        Dataset to store
    chunks : tuple[int, ...] | None
        Chunk size in each dimension
    compression : int | str | bool | None
        Compression to use for the dataset
    external : Iterable[tuple[str, int, int]] | None
        List of tuples to be passed to h5p.set_external
    track_order : bool | None
        If True, set tracking and indexing of creation order for object attributes;
        otherwise, dcpl.set_attr_creation_order(0) is called
    dcpl : h5p.PropDCID | None
        Dataset Creation Property List to use; if unspecified, an new dcpl is created

    Returns
    -------
    np.ndarray | None
        Data used to create the dataset
    """
    # Convert data to a C-contiguous ndarray
    if data is not None and not isinstance(data, Empty):
        # normalize strings -> np.dtype objects
        if dtype is not None:
            _dtype = np.dtype(dtype)
        else:
            _dtype = None

        # if we are going to a f2 datatype, pre-convert in python
        # to workaround a possible h5py bug in the conversion.
        is_small_float = (
            _dtype is not None and _dtype.kind == "f" and _dtype.itemsize == 2
        )
        data = np.asarray(
            data, order="C", dtype=(_dtype if is_small_float else guess_dtype(data))
        )

    # Validate shape
    if shape is None:
        if data is None:
            if dtype is None:
                raise TypeError("One of data, shape or dtype must be specified.")
            data = Empty(dtype)
        shape = data.shape
    else:
        shape = (shape,) if isinstance(shape, int) else tuple(shape)
        if data is not None and (
            np.prod(shape, dtype=np.ulonglong)
            != np.prod(data.shape, dtype=np.ulonglong)
        ):
            raise ValueError(
                f"Shape tuple {shape} is incompatible with data shape {data.shape}"
            )

    if isinstance(maxshape, int):
        maxshape = (maxshape,)

    # Validate chunk shape
    if isinstance(chunks, int) and not isinstance(chunks, bool):
        chunks = (chunks,)

    # The original make_new_dset errors here if the shape is less than the
    # chunk size, but we avoid doing that as we cannot change the chunk size
    # for a dataset for any version once it is created. See #34.

    if isinstance(dtype, Datatype):
        # Named types are used as-is
        tid = dtype.id
        dtype = tid.dtype  # Following code needs this
    else:
        # Validate dtype
        if dtype is None and data is None:
            dtype = np.dtype("=f4")
        elif dtype is None and data is not None:
            dtype = data.dtype
        else:
            dtype = np.dtype(dtype)
        tid = h5t.py_create(dtype, logical=1)

    # Legacy
    if (
        any((compression, shuffle, fletcher32, maxshape, scaleoffset))
        and chunks is False
    ):
        raise ValueError(
            "Chunked format required for given storage options:\n"
            + textwrap.indent(
                "\n".join(
                    [
                        f"compression: {compression}",
                        f"shuffle: {shuffle}",
                        f"fletcher32: {fletcher32}",
                        f"maxshape: {maxshape}",
                        f"scaleoffset: {scaleoffset}",
                    ]
                ),
                "  ",
            )
        )

    # Legacy
    if compression is True:
        if compression_opts is None:
            compression_opts = 4
        compression = "gzip"

    # Legacy
    if compression in _LEGACY_GZIP_COMPRESSION_VALS:
        if compression_opts is not None:
            raise TypeError(
                "Conflict in compression options; "
                f"if compression is one of {_LEGACY_GZIP_COMPRESSION_VALS},"
                f"compression_opts must be None.\n"
                + textwrap.indent(
                    "\n".join(
                        f"compression: {compression}",
                        f"compression_opts: {compression_opts}",
                    )
                )
            )
        compression_opts = compression
        compression = "gzip"
    dcpl = filters.fill_dcpl(
        dcpl or h5p.create(h5p.DATASET_CREATE),
        shape,
        dtype,
        chunks,
        compression,
        compression_opts,
        shuffle,
        fletcher32,
        maxshape,
        scaleoffset,
        external,
    )

    if fillvalue is not None:
        fillvalue = np.array(fillvalue)
        dcpl.set_fill_value(fillvalue)

    if track_times in (True, False):
        dcpl.set_obj_track_times(track_times)
    elif track_times is not None:
        raise TypeError("track_times must be either True or False")
    if track_order == True:
        dcpl.set_attr_creation_order(h5p.CRT_ORDER_TRACKED | h5p.CRT_ORDER_INDEXED)
    elif track_order == False:
        dcpl.set_attr_creation_order(0)
    elif track_order is not None:
        raise TypeError("track_order must be either True or False")

    if maxshape is not None:
        maxshape = tuple(m if m is not None else h5s.UNLIMITED for m in maxshape)

    if isinstance(data, Empty):
        raise NotImplementedError("Empty datasets are not supported.")
    return data


class InMemoryDataset(Dataset):
    """
    Class that looks like a h5py.Dataset but is backed by a versioned dataset

    The versioned dataset can be modified, which performs modifications
    in-memory only.
    """

    def __init__(self, bind, parent, **kwargs):
        # Hold a reference to the original bind so h5py doesn't invalidate the id
        # XXX: We need to handle deallocation here properly when our object
        # gets deleted or closed.
        self.orig_bind = bind
        super().__init__(InMemoryDatasetID(bind.id), **kwargs)
        self._parent = parent
        self._attrs = dict(super().attrs)

    def __repr__(self):
        name = posixpath.basename(posixpath.normpath(self.name))
        namestr = '"%s"' % (name if name != "" else "/")
        return '<%s %s: shape %s, type "%s">' % (
            self.__class__.__name__,
            namestr,
            self.shape,
            self.dtype.str,
        )

    @property
    def data_dict(self):
        return self.id.data_dict

    @property
    def compression(self):
        name = self.name
        if self.parent.name in name:
            name = name[len(self.parent.name) + 1 :]
        return self.parent.compression[name]

    @compression.setter
    def compression(self, value):
        self.parent.set_compression(self.item, value)

    @property
    def compression_opts(self):
        name = self.name
        if self.parent.name in name:
            name = name[len(self.parent.name) + 1 :]
        return self.parent.compression_opts[self.name]

    @compression_opts.setter
    def compression_opts(self, value):
        self.parent.set_compression_opts(self.name, value)

    def as_dtype(self, name, dtype, parent, casting="unsafe"):
        """
        Return a copy of `self` as a new dataset with the given `name` and `dtype`
        in the group `parent`.

        `casting` should be as in the numpy astype() method.

        """
        if self.fillvalue is not None:
            new_fillvalue = self.fillvalue.astype(dtype, casting=casting)
        else:
            new_fillvalue = None
        new_data_dict = {}
        for c, index in self.data_dict.copy().items():
            if isinstance(index, Slice):
                self[c.raw]
                assert not isinstance(self.data_dict[c], Slice)
            new_data_dict[c] = self.data_dict[c].astype(dtype, casting=casting)
        new_dataset = InMemorySparseDataset.from_data_dict(
            name,
            new_data_dict,
            shape=self.shape,
            dtype=dtype,
            parent=parent,
            chunks=self.chunks,
            fillvalue=new_fillvalue,
        )
        parent[name] = new_dataset
        return new_dataset

    @property
    def fillvalue(self):
        if super().fillvalue is not None:
            return super().fillvalue
        if self.dtype.metadata:
            # Custom h5py string dtype. Make sure to use a fillvalue of ''
            if "vlen" in self.dtype.metadata:
                # h5py 3 reads str variable length datasets as bytes. See
                # https://docs.h5py.org/en/stable/whatsnew/3.0.html#breaking-changes-deprecations.
                if h5py_version.startswith("3") and self.dtype.metadata["vlen"] == str:
                    return bytes()
                return self.dtype.metadata["vlen"]()
            elif "h5py_encoding" in self.dtype.metadata:
                return self.dtype.type()
        return np.zeros((), dtype=self.dtype)[()]

    @property
    def chunks(self):
        return ChunkSize(self.id.chunks)

    @property
    def attrs(self):
        return self._attrs

    @property
    def parent(self):
        return self._parent

    def __array__(self, dtype=None):
        return self.__getitem__(())

    def resize(self, size, axis=None):
        """Resize the dataset, or the specified axis.

        The rank of the dataset cannot be changed.

        "Size" should be a shape tuple, or if an axis is specified, an integer.

        BEWARE: This functions differently than the NumPy resize() method!
        The data is not "reshuffled" to fit in the new shape; each axis is
        grown or shrunk independently.  The coordinates of existing data are
        fixed.
        """
        self.parent._check_committed()
        # This boilerplate code is based on h5py.Dataset.resize
        if axis is not None:
            if not (axis >= 0 and axis < self.id.rank):
                raise ValueError("Invalid axis (0 to %s allowed)" % (self.id.rank - 1))
            try:
                newlen = int(size)
            except TypeError:
                raise TypeError("Argument must be a single int if axis is specified")
            size = list(self.shape)
            size[axis] = newlen

        size = tuple(size)
        # === END CODE FROM h5py.Dataset.resize ===

        old_shape = self.shape
        data_dict = self.id.data_dict
        can_read_direct = self.id.can_read_direct

        old_shape_idx = Tuple(*[Slice(0, i) for i in old_shape])

        new_data_dict = {}
        # Read entire Dataset into in-memory array in one operation.
        # Usually that's much faster than reading individual chunks.
        old_data = self.get_index(old_shape_idx.raw, can_read_direct=can_read_direct)
        # Now read chunks out of the array.
        for chunk, _, index_raw in as_subchunk_map(old_shape_idx, size, self.chunks):
            if chunk in data_dict:
                new_data_dict[chunk] = data_dict[chunk]
            else:
                data = np.full(chunk.newshape(size), self.fillvalue, dtype=self.dtype)
                data[index_raw] = old_data[chunk.raw]
                new_data_dict[chunk] = data

        self.id.data_dict = new_data_dict
        self.id.shape = size

    @with_phil
    def get_index(
        self,
        args: slice | Slice | Tuple | tuple | h5r.RegionReference,
        new_dtype: str | None = None,
        can_read_direct: bool | None = None,
    ) -> np.ndarray:
        """Read a slice from the HDF5 dataset given by the index.

        Takes slices and recarray-style field names (more than one is
        allowed!) in any order.  Obeys basic NumPy rules, including
        broadcasting.

        Parameters
        ----------
        args : slice | Slice | Tuple | tuple | h5r.RegionReference
            Index to read from the Dataset
        new_dtype : str | None
            Dtype of the returned array
        can_read_direct : bool | None
            True if we can read directly from the underlying hdf5 Dataset, False otherwise.
            This should be the value of the InMemoryDatasetID instance's ``can_read_direct``
            property for this Dataset.

            If None, ``self.id.can_read_direct`` is evaluated first to determine if data can
            be read directly from the underlying dataset.

        Returns
        -------
        np.ndarray
            Array containing data from this dataset from the requested index
        """
        # This boilerplate code is based on h5py.Dataset.__getitem__
        args = args if isinstance(args, tuple) else (args,)

        # Sort field names from the rest of the args.
        names = tuple(x for x in args if isinstance(x, str))

        if names:
            # Read a subset of the fields in this structured dtype
            if len(names) == 1:
                names = names[0]  # Read with simpler dtype of this field
            args = tuple(x for x in args if not isinstance(x, str))
            return self.fields(names, _prior_dtype=new_dtype)[args]

        if new_dtype is None:
            new_dtype = self.dtype
        mtype = h5t.py_create(new_dtype)

        # === Special-case region references ====

        if len(args) == 1 and isinstance(args[0], h5r.RegionReference):
            obj = h5r.dereference(args[0], self.id)
            if obj != self.id:
                raise ValueError("Region reference must point to this dataset")

            sid = h5r.get_region(args[0], self.id)
            mshape = guess_shape(sid)
            if mshape is None:
                # 0D with no data (NULL or deselected SCALAR)
                return Empty(new_dtype)
            out = np.empty(mshape, dtype=new_dtype)
            if out.size == 0:
                return out

            sid_out = h5s.create_simple(mshape)
            sid_out.select_all()
            self.id.read(sid_out, sid, out, mtype)
            return out

        # === END CODE FROM h5py.Dataset.__getitem__ ===

        idx = ndindex(args).expand(self.shape)

        if can_read_direct is None:
            can_read_direct = self.id.can_read_direct

        if can_read_direct:
            return super().__getitem__(idx.raw)

        arr = np.ndarray(idx.newshape(self.shape), new_dtype, order="C")

        for chunk, arr_idx_raw, index_raw in as_subchunk_map(
            idx, self.shape, self.chunks
        ):
            if chunk not in self.id.data_dict:
                self.id.data_dict[chunk] = np.broadcast_to(
                    self.fillvalue, chunk.newshape(self.shape)
                )
            elif isinstance(self.id.data_dict[chunk], Slice):
                raw_idx = tuple(
                    [self.id.data_dict[chunk].raw]
                    + [slice(0, len(i)) for i in chunk.args[1:]]
                )
                self.id.data_dict[chunk] = self.id._read_chunk(raw_idx)

            if self.id.data_dict[chunk].size != 0:
                arr[arr_idx_raw] = self.id.data_dict[chunk][index_raw]

        # Return arr as a scalar if it is shape () (matching h5py)
        return arr[()]

    @with_phil
    def __getitem__(
        self,
        args: slice | Slice | Tuple | tuple | h5r.RegionReference,
    ) -> np.ndarray:
        return self.get_index(args)

    @with_phil
    def __setitem__(self, args, val):
        """Write to the HDF5 dataset from a NumPy array.

        NumPy's broadcasting rules are honored, for "simple" indexing
        (slices and integers).  For advanced indexing, the shapes must
        match.
        """
        self.parent._check_committed()
        # This boilerplate code is based on h5py.Dataset.__setitem__
        args = args if isinstance(args, tuple) else (args,)

        # Sort field indices from the slicing
        names = tuple(x for x in args if isinstance(x, str))
        args = tuple(x for x in args if not isinstance(x, str))

        # Generally we try to avoid converting the arrays on the Python
        # side.  However, for compound literals this is unavoidable.
        vlen = h5t.check_vlen_dtype(self.dtype)
        if vlen is not None and vlen not in (bytes, str):
            try:
                val = np.asarray(val, dtype=vlen)
            except ValueError:
                try:
                    val = np.array(
                        [np.array(x, dtype=vlen) for x in val], dtype=self.dtype
                    )
                except ValueError:
                    pass
            if vlen == val.dtype:
                if val.ndim > 1:
                    tmp = np.empty(shape=val.shape[:-1], dtype=object)
                    tmp.ravel()[:] = [
                        i
                        for i in val.reshape(
                            (
                                np.prod(val.shape[:-1], dtype=np.ulonglong),
                                val.shape[-1],
                            )
                        )
                    ]
                else:
                    tmp = np.array([None], dtype=object)
                    tmp[0] = val
                val = tmp
        elif self.dtype.kind == "O" or (
            self.dtype.kind == "V"
            and (not isinstance(val, np.ndarray) or val.dtype.kind != "V")
            and (self.dtype.subdtype == None)
        ):
            if len(names) == 1 and self.dtype.fields is not None:
                # Single field selected for write, from a non-array source
                if not names[0] in self.dtype.fields:
                    raise ValueError("No such field for indexing: %s" % names[0])
                dtype = self.dtype.fields[names[0]][0]
                cast_compound = True
            else:
                dtype = self.dtype
                cast_compound = False

            val = np.asarray(val, dtype=dtype.base, order="C")
            if cast_compound:
                val = val.view(np.dtype([(names[0], dtype)]))
                val = val.reshape(val.shape[: len(val.shape) - len(dtype.shape)])
        else:
            val = np.asarray(val, order="C")

        # Check for array dtype compatibility and convert
        if self.dtype.subdtype is not None:
            shp = self.dtype.subdtype[1]
            valshp = val.shape[-len(shp) :]
            if valshp != shp:  # Last dimension has to match
                raise TypeError(
                    "When writing to array types, last N dimensions have to match (got %s, but should be %s)"
                    % (
                        valshp,
                        shp,
                    )
                )
            mtype = h5t.py_create(np.dtype((val.dtype, shp)))
            # mshape = val.shape[0:len(val.shape)-len(shp)]

        # Make a compound memory type if field-name slicing is required
        elif len(names) != 0:
            # mshape = val.shape

            # Catch common errors
            if self.dtype.fields is None:
                raise TypeError("Illegal slicing argument (not a compound dataset)")
            mismatch = [x for x in names if x not in self.dtype.fields]
            if len(mismatch) != 0:
                mismatch = ", ".join('"%s"' % x for x in mismatch)
                raise ValueError(
                    "Illegal slicing argument (fields %s not in dataset type)"
                    % mismatch
                )

            # Write non-compound source into a single dataset field
            if len(names) == 1 and val.dtype.fields is None:
                subtype = h5t.py_create(val.dtype)
                mtype = h5t.create(h5t.COMPOUND, subtype.get_size())
                mtype.insert(self._e(names[0]), 0, subtype)

            # Make a new source type keeping only the requested fields
            else:
                fieldnames = [
                    x for x in val.dtype.names if x in names
                ]  # Keep source order
                mtype = h5t.create(h5t.COMPOUND, val.dtype.itemsize)
                for fieldname in fieldnames:
                    subtype = h5t.py_create(val.dtype.fields[fieldname][0])
                    offset = val.dtype.fields[fieldname][1]
                    mtype.insert(self._e(fieldname), offset, subtype)

        # Use mtype derived from array (let DatasetID.write figure it out)
        else:
            mtype = None

        # === END CODE FROM h5py.Dataset.__setitem__ ===

        idx = ndindex(args).reduce(self.shape)

        val = np.broadcast_to(val, idx.newshape(self.shape))

        for c, val_idx_raw, index_raw in as_subchunk_map(idx, self.shape, self.chunks):
            if c not in self.id.data_dict:
                # Broadcasted arrays do not actually consume memory
                fill = np.broadcast_to(self.fillvalue, c.newshape(self.shape))
                self.id.data_dict[c] = fill.astype(self.dtype)
            elif isinstance(self.id.data_dict[c], Slice):
                raw_idx = tuple(
                    [self.id.data_dict[c].raw] + [slice(0, len(i)) for i in c.args[1:]]
                )
                self.id.data_dict[c] = self.id._read_chunk(raw_idx)

            if self.id.data_dict[c].size != 0:
                if not self.id.data_dict[c].flags.writeable:
                    # self.id.data_dict[c] is a broadcasted array from above
                    self.id.data_dict[c] = self.id.data_dict[c].copy()
                self.id.data_dict[c][index_raw] = val[val_idx_raw]


class DatasetLike:
    """
    Superclass for classes that look like h5py.Dataset

    Subclasses should have the following properties defined (properties
    starting with an underscore will be computed if they are None)

    name
    shape
    dtype
    _fillvalue
    parent (the parent group)
    """

    @property
    def size(self):
        return np.prod(self.shape)

    @property
    def fillvalue(self):
        if self._fillvalue is not None:
            return np.array([self._fillvalue], dtype=self.dtype)[0]
        if self.dtype.metadata:
            # Custom h5py string dtype. Make sure to use a fillvalue of ''
            if "vlen" in self.dtype.metadata:
                # h5py 3 reads str variable length datasets as bytes. See
                # https://docs.h5py.org/en/stable/whatsnew/3.0.html#breaking-changes-deprecations.
                if h5py_version.startswith("3") and self.dtype.metadata["vlen"] == str:
                    return bytes()
                return self.dtype.metadata["vlen"]()
            elif "h5py_encoding" in self.dtype.metadata:
                return b""
        return np.zeros((), dtype=self.dtype)[()]

    @property
    def ndim(self):
        return len(self.shape)

    def __bool__(self):
        return bool(self.size)

    def __len__(self):
        return self.len()

    def len(self):
        """
        Length of the first axis
        """
        shape = self.shape
        if len(shape) == 0:
            raise TypeError("Attempt to take len() of scalar dataset")
        return shape[0]

    def __repr__(self):
        name = posixpath.basename(posixpath.normpath(self.name))
        namestr = '"%s"' % (name if name != "" else "/")
        return '<%s %s: shape %s, type "%s">' % (
            self.__class__.__name__,
            namestr,
            self.shape,
            self.dtype.str,
        )

    def __iter__(self):
        """Iterate over the first axis. TypeError if scalar.

        BEWARE: Modifications to the yielded data are *NOT* written to file.
        """
        shape = self.shape
        if len(shape) == 0:
            raise TypeError("Can't iterate over a scalar dataset")
        for i in range(shape[0]):
            yield self[i]

    @property
    def compression(self):
        name = self.name
        if self.parent.name in name:
            name = name[len(self.parent.name) + 1 :]
        return self.parent.compression[name]

    @compression.setter
    def compression(self, value):
        name = self.name
        if self.parent.name in name:
            name = name[len(self.parent.name) + 1 :]
        self.parent.set_compression(name, value)

    @property
    def compression_opts(self):
        name = self.name
        if self.parent.name in name:
            name = name[len(self.parent.name) + 1 :]
        return self.parent.compression_opts[name]

    @compression_opts.setter
    def compression_opts(self, value):
        name = self.name
        if self.parent.name in name:
            name = name[len(self.parent.name) + 1 :]
        self.parent.set_compression_opts(name, value)


class InMemoryArrayDataset(DatasetLike):
    """
    Class that looks like a h5py.Dataset but is backed by an array
    """

    def __init__(self, name, array, parent, fillvalue=None, chunks=None):
        self.name = name
        self._array = array
        self._dtype = None
        self.attrs = {}
        self.parent = parent
        self._fillvalue = fillvalue
        if chunks is None:
            chunks = parent.chunks[name]
        self._chunks = chunks

    def as_dtype(self, name, dtype, parent, casting="unsafe"):
        """
        Return a copy of `self` as a new dataset with the given `name` and `dtype`
        in the group `parent`.

        `casting` should be as in the numpy astype() method.

        """
        return self.__class__(
            name, self.array.astype(dtype, casting=casting), parent=parent
        )

    @property
    def array(self):
        return self._array

    @array.setter
    def array(self, array):
        self._array = array

    @property
    def shape(self):
        return self._array.shape

    @property
    def dtype(self):
        if self._dtype is not None:
            return self._dtype
        return self._array.dtype

    def __getitem__(self, item):
        return self.array.__getitem__(item)

    def __setitem__(self, item, value):
        self.parent._check_committed()
        self.array.__setitem__(item, value)

    def __array__(self, dtype=None):
        return self.array

    @property
    def chunks(self):
        return self._chunks

    def resize(self, size, axis=None):
        self.parent._check_committed()
        if axis is not None:
            if not (axis >= 0 and axis < self.ndim):
                raise ValueError("Invalid axis (0 to %s allowed)" % (self.ndim - 1))
            try:
                newlen = int(size)
            except TypeError:
                raise TypeError("Argument must be a single int if axis is specified")
            size = list(self.shape)
            size[axis] = newlen

        old_shape = self.shape
        size = tuple(size)
        if all(new <= old for new, old in zip(size, old_shape)):
            # Don't create a new array if the old one can just be sliced in
            # memory.
            idx = tuple(slice(0, i) for i in size)
            self.array = self.array[idx]
        else:
            old_shape_idx = Tuple(*[Slice(0, i) for i in old_shape])
            new_shape_idx = Tuple(*[Slice(0, i) for i in size])
            new_array = np.full(size, self.fillvalue, dtype=self.dtype)
            new_array[old_shape_idx.as_subindex(new_shape_idx).raw] = self.array[
                new_shape_idx.as_subindex(old_shape_idx).raw
            ]
            self.array = new_array


class InMemorySparseDataset(DatasetLike):
    """
    Class that looks like a Dataset that has no data (only the fillvalue)
    """

    def __init__(self, name, *, shape, dtype, parent, chunks=None, fillvalue=None):
        if shape is None:
            raise TypeError("shape must be specified for sparse datasets")
        self.name = name
        self.shape = shape
        self.dtype = np.dtype(dtype)
        self.attrs = {}
        self._fillvalue = fillvalue
        if chunks in [True, None]:
            if len(shape) == 1:
                chunks = (DEFAULT_CHUNK_SIZE,)
            else:
                raise NotImplementedError(
                    "chunks must be specified for multi-dimensional datasets"
                )
        self.chunks = ChunkSize(chunks)
        self.parent = parent

        # This works like a mix between InMemoryArrayDataset and
        # InMemoryDatasetID. Explicit array data is stored in a data_dict like
        # with InMemoryDatasetID, but unlike it, missing data (which equals
        # the fill value) is omitted.
        self.data_dict = {}

    def as_dtype(self, name, dtype, parent, casting="unsafe"):
        """
        Return a copy of `self` as a new dataset with the given `name` and `dtype`
        in the group `parent`.

        `casting` should be as in the numpy astype() method.

        """
        if self.fillvalue is not None:
            new_fillvalue = self.fillvalue.astype(dtype, casting=casting)
        else:
            new_fillvalue = None
        new_data_dict = {}
        for c, index in self.data_dict.copy().items():
            new_data_dict[c] = self.data_dict[c].astype(dtype, casting=casting)

        return self.from_data_dict(
            name,
            new_data_dict,
            dtype=dtype,
            parent=parent,
            fillvalue=new_fillvalue,
            chunks=self.chunks,
            shape=self.shape,
        )

    @classmethod
    def from_data_dict(
        cls, name, data_dict, *, shape, dtype, parent, chunks, fillvalue=None
    ):
        """
        Create a InMemorySparseDataset from a data dict.

        This does not do any consistency checks with the metadata provide.
        """
        dataset = cls(
            name,
            shape=shape,
            dtype=dtype,
            parent=parent,
            chunks=chunks,
            fillvalue=fillvalue,
        )
        dataset.data_dict = data_dict
        return dataset

    @classmethod
    def from_dataset(cls, dataset, parent=None):
        # np.testing.assert_equal(dataset[()], dataset.fillvalue)
        return cls(
            dataset.name,
            shape=dataset.shape,
            dtype=dataset.dtype,
            parent=parent or dataset.parent,
            chunks=dataset.chunks,
            fillvalue=dataset.fillvalue,
        )

    def resize(self, size, axis=None):
        if axis is not None:
            if not (axis >= 0 and axis < self.ndim):
                raise ValueError("Invalid axis (0 to %s allowed)" % (self.ndim - 1))
            try:
                newlen = int(size)
            except TypeError:
                raise TypeError("Argument must be a single int if axis is specified")
            size = list(self.shape)
            size[axis] = newlen

        size = tuple(size)
        if len(size) > 1:
            raise NotImplementedError("More than one dimension is not yet supported")
        self.shape = size

    def __getitem__(self, index):
        idx = ndindex(index).reduce(self.shape)

        newshape = idx.newshape(self.shape)
        arr = np.full(newshape, self.fillvalue, dtype=self.dtype)

        for c, arr_idx_raw, chunk_idx_raw in as_subchunk_map(
            idx, self.shape, self.chunks
        ):
            if c not in self.data_dict:
                fill = np.broadcast_to(self.fillvalue, c.newshape(self.shape))
                self.data_dict[c] = fill

            if self.data_dict[c].size != 0:
                arr[arr_idx_raw] = self.data_dict[c][chunk_idx_raw]

        # Return arr as a scalar if it is shape () (matching h5py)
        return arr[()]

    def __setitem__(self, index, value):
        self.parent._check_committed()

        idx = ndindex(index).reduce(self.shape)

        val = np.broadcast_to(value, idx.newshape(self.shape))

        for c, val_idx_raw, chunk_idx_raw in as_subchunk_map(
            idx, self.shape, self.chunks
        ):
            if c not in self.data_dict:
                # Broadcasted arrays do not actually consume memory
                fill = np.broadcast_to(self.fillvalue, c.newshape(self.shape))
                self.data_dict[c] = fill

            if self.data_dict[c].size != 0:
                if not self.data_dict[c].flags.writeable:
                    # self.data_dict[c] is a broadcasted array from above
                    self.data_dict[c] = self.data_dict[c].copy()
                self.data_dict[c][chunk_idx_raw] = val[val_idx_raw]


class DatasetWrapper(DatasetLike):
    def __init__(self, dataset):
        self.dataset = dataset

    def __getattr__(self, attr):
        return getattr(self.dataset, attr)

    def __setitem__(self, index, value):
        if isinstance(self.dataset, InMemoryDataset) and ndindex(index).expand(
            self.shape
        ) == Tuple().expand(self.shape):
            new_dataset = InMemoryArrayDataset(
                self.name,
                np.broadcast_to(value, self.shape).astype(self.dtype),
                self.parent,
                fillvalue=self.fillvalue,
                chunks=self.chunks,
            )
            new_dataset.attrs = self.dataset.attrs
            self.dataset = new_dataset
            return
        self.dataset.__setitem__(index, value)

    def __getitem__(self, index):
        return self.dataset.__getitem__(index)


class InMemoryDatasetID(h5d.DatasetID):
    def __init__(self, _id):
        # super __init__ is handled by DatasetID.__cinit__ automatically
        self._data_dict = None
        with phil:
            sid = self.get_space()
            self._shape = sid.get_simple_extent_dims()
        self._reshaped = False

        attr = h5a.open(self, b"raw_data")
        htype = h5t.py_create(attr.dtype)
        _arr = np.ndarray(attr.shape, dtype=attr.dtype, order="C")
        attr.read(_arr, mtype=htype)
        raw_data_name = _arr[()]
        if isinstance(raw_data_name, bytes):
            raw_data_name = raw_data_name.decode("utf-8")

        fid = h5i.get_file_id(self)
        g = Group(fid)
        self.raw_data = g[raw_data_name]
        self.chunks = tuple(self.raw_data.attrs["chunks"])

        fillvalue_a = np.empty((1,), dtype=self.dtype)
        dcpl = self.get_create_plist()
        dcpl.get_fill_value(fillvalue_a)
        self.fillvalue = fillvalue_a[0]

    @property
    def data_dict(self):
        if self._data_dict is None:
            dcpl = self.get_create_plist()

            is_virtual: bool = dcpl.get_layout() == h5d.VIRTUAL

            if not is_virtual:
                # A dataset created with only a fillvalue will be nonvirtual,
                # since create_virtual_dataset makes a nonvirtual dataset when
                # there are no virtual sources.
                self._data_dict = {}
            # Same as dataset.get_virtual_sources
            elif 0 in self.shape:
                # Work around https://github.com/h5py/h5py/issues/1660
                empty_idx = Tuple().expand(self.shape)
                self._data_dict = {empty_idx: Slice()}
            else:
                self._data_dict = build_data_dict(dcpl, self.raw_data.name)

        return self._data_dict

    @data_dict.setter
    def data_dict(self, value):
        self._data_dict = value

    @property
    def can_read_direct(self) -> bool:
        """Check whether reading directly from the underlying dataset is okay.

        Returns
        -------
        bool
            If this is True, then h5py.Dataset.__getitem__ can be used, which may
            be faster than InMemoryDataset.__getitem__. This will happen in
            particular when reading a read-only dataset.
        """
        if self._data_dict is not None and any(
            isinstance(i, np.ndarray) for i in self._data_dict.values()
        ):
            return False
        if self._reshaped:
            return False
        return True

    def set_extent(self, shape):
        raise NotImplementedError("Resizing an InMemoryDataset other than via resize()")

    @property
    def shape(self):
        return self._shape

    @shape.setter
    def shape(self, size):
        self._shape = size
        self._reshaped = True

    def _read_chunk(self, chunk_idx):
        return self.raw_data[chunk_idx]

    def write(self, mspace, fspace, arr_obj, mtype=None, dxpl=None):
        raise NotImplementedError(
            "Writing to an InMemoryDataset other than via __setitem__"
        )



from __future__ import annotations

import logging
import os
import textwrap
from collections.abc import Iterator

import numpy as np
from h5py import Dataset, VirtualLayout, VirtualSource, h5s, h5z
from h5py._hl.filters import guess_chunk
from ndindex import ChunkSize, Slice, Tuple, ndindex
from numpy.testing import assert_array_equal

from .hashtable import Hashtable

DEFAULT_CHUNK_SIZE = 2**12
DATA_VERSION = 4
# data_version 2 has broken hashtables, always need to rebuild
# data_version 3 hash collisions for string arrays which, when concatenated, give the same string
CORRUPT_DATA_VERSIONS = frozenset([2, 3])


def normalize_dtype(dtype):
    return np.array([], dtype=dtype).dtype


def get_chunks(shape, dtype, chunk_size):
    # TODO: Implement this
    if len(shape) > 1:
        raise NotImplementedError
    return (chunk_size,)


def initialize(f):
    import datetime

    from .versions import TIMESTAMP_FMT

    version_data = f.create_group("_version_data")
    versions = version_data.create_group("versions")
    versions.create_group("__first_version__")
    versions.attrs["current_version"] = "__first_version__"
    ts = datetime.datetime.now(datetime.timezone.utc)
    versions["__first_version__"].attrs["timestamp"] = ts.strftime(TIMESTAMP_FMT)
    versions.attrs["data_version"] = DATA_VERSION


def create_base_dataset(
    f,
    name,
    *,
    shape=None,
    data=None,
    dtype=None,
    chunks=True,
    compression=None,
    compression_opts=None,
    fillvalue=None,
):
    # Validate shape (based on h5py._hl.dataset.make_new_dset)
    if shape is None:
        if data is None:
            if dtype is None:
                raise TypeError("One of data, shape or dtype must be specified")
            raise NotImplementedError("empty datasets are not yet implemented")
        shape = data.shape
    else:
        shape = (shape,) if isinstance(shape, int) else tuple(shape)
        if data is not None and (
            np.prod(shape, dtype=np.ulonglong)
            != np.prod(data.shape, dtype=np.ulonglong)
        ):
            raise ValueError("Shape tuple is incompatible with data")

    ndims = len(shape)
    if isinstance(chunks, int) and not isinstance(chunks, bool):
        chunks = (chunks,)
    if chunks in [True, None]:
        if ndims == 0:
            # Chunks are not allowed for scalar datasets; keeping original
            # behavior here
            chunks = (DEFAULT_CHUNK_SIZE,)
        elif ndims == 1:
            chunks = guess_chunk(shape, None, data.dtype.itemsize)
        else:
            raise NotImplementedError(
                "chunks must be specified for multi-dimensional datasets"
            )
    group = f["_version_data"].create_group(name)

    if dtype is None:
        # https://github.com/h5py/h5py/issues/1474
        dtype = data.dtype
    dtype = normalize_dtype(dtype)
    if dtype.metadata and (
        "vlen" in dtype.metadata or "h5py_encoding" in dtype.metadata
    ):
        # h5py string dtype
        # (https://h5py.readthedocs.io/en/2.10.0/strings.html). Setting the
        # fillvalue in this case doesn't work
        # (https://github.com/h5py/h5py/issues/941).
        if fillvalue not in [0, "", b"", None]:
            raise ValueError(
                "Non-default fillvalue not supported for variable length strings"
            )
        fillvalue = None
    dataset = group.create_dataset(
        "raw_data",
        shape=(0,) + chunks[1:],
        chunks=tuple(chunks),
        maxshape=(None,) + chunks[1:],
        dtype=dtype,
        compression=compression,
        compression_opts=compression_opts,
        fillvalue=fillvalue,
    )
    dataset.attrs["chunks"] = chunks
    return write_dataset(f, name, data, chunks=chunks)


def write_dataset(
    f,
    name,
    data,
    chunks=None,
    dtype=None,
    compression=None,
    compression_opts=None,
    fillvalue=None,
):
    if name not in f["_version_data"]:
        return create_base_dataset(
            f,
            name,
            data=data,
            dtype=dtype,
            chunks=chunks,
            compression=compression,
            compression_opts=compression_opts,
            fillvalue=fillvalue,
        )

    ds = f["_version_data"][name]["raw_data"]
    if isinstance(chunks, int) and not isinstance(chunks, bool):
        chunks = (chunks,)
    if chunks is None:
        chunks = tuple(ds.attrs["chunks"])
    else:
        if chunks != tuple(ds.attrs["chunks"]):
            raise ValueError(
                "Chunk size specified but doesn't match already existing chunk size"
            )

    if dtype is not None:
        if dtype != ds.dtype:
            raise ValueError("dtype specified but doesn't match already existing dtype")

    if (
        compression
        and compression not in ds._filters
        or compression_opts
        and compression_opts != ds._filters[ds.compression]
    ):
        available_filters = textwrap.indent(
            "\n".join(str(filter) for filter in get_available_filters()), "  "
        )
        raise ValueError(
            "Compression options can only be specified for the first version of a dataset.\n"
            f"Dataset: {name}\n"
            f"Current filters: {ds._filters}\n"
            f"Available hdf5 compression types:\n{available_filters}"
        )
    if fillvalue is not None and fillvalue != ds.fillvalue:
        dtype = ds.dtype
        if dtype.metadata and (
            "vlen" in dtype.metadata or "h5py_encoding" in dtype.metadata
        ):
            # Variable length string dtype. The ds.fillvalue will be None in
            # this case (see create_virtual_dataset() below)
            pass
        else:
            raise ValueError(f"fillvalues do not match ({fillvalue} != {ds.fillvalue})")
    if data.dtype != ds.dtype:
        raise ValueError(f"dtypes do not match ({data.dtype} != {ds.dtype})")
    # TODO: Handle more than one dimension
    old_shape = ds.shape
    slices = {}
    slices_to_write = {}
    chunk_size = chunks[0]

    validate_reused_chunks = os.environ.get(
        "ENABLE_CHUNK_REUSE_VALIDATION", "false"
    ).lower() in ("1", "true")

    with Hashtable(f, name) as hashtable:
        old_chunks = hashtable.largest_index
        chunks_reused = 0

        if len(data.shape) != 0:
            for data_slice in ChunkSize(chunks).indices(data.shape):
                data_s = data[data_slice.raw]
                data_hash = hashtable.hash(data_s)

                if data_hash in hashtable:
                    hashed_slice = hashtable[data_hash]
                    slices[data_slice] = hashed_slice

                    if validate_reused_chunks:
                        _verify_new_chunk_reuse(
                            raw_dataset=ds,
                            new_data=data,
                            data_hash=data_hash,
                            hashed_slice=hashed_slice,
                            chunk_being_written=data_s,
                            slices_to_write=slices_to_write,
                        )

                    chunks_reused += 1

                else:
                    idx = hashtable.largest_index
                    raw_slice = Slice(
                        idx * chunk_size, idx * chunk_size + data_s.shape[0]
                    )
                    slices[data_slice] = raw_slice
                    hashtable[data_hash] = raw_slice
                    slices_to_write[raw_slice] = data_slice

            ds.resize((old_shape[0] + len(slices_to_write) * chunk_size,) + chunks[1:])
            for raw_slice, data_slice in slices_to_write.items():
                data_s = data[data_slice.raw]
                idx = Tuple(raw_slice, *[slice(0, i) for i in data_s.shape[1:]])
                ds[idx.raw] = data[data_slice.raw]

        new_chunks = hashtable.largest_index

    logging.debug(
        "  %s: " "New chunks written: %d; " "Number of chunks reused: %d",
        name,
        new_chunks - old_chunks,
        chunks_reused,
    )

    return slices


def _verify_new_chunk_reuse(
    raw_dataset: Dataset,
    new_data: np.ndarray,
    data_hash: bytes,
    hashed_slice: Slice,
    chunk_being_written: np.ndarray,
    slices_to_write: dict[Slice, Tuple] | None = None,
    data_to_write: dict[Slice, np.ndarray] | None = None,
) -> None:
    """Check that the data from the hashed slice matches the data to be written.

    Raises a ValueError if the data reference by the hashed slice doesn't match the
    underlying raw data.

    This function retrieves a reused chunk of data either from the ``slices_to_write``,
    if the data has not yet been written to the file, or from the ``raw_data`` that has
    already been written.

    Parameters
    ----------
    raw_dataset : Dataset
        Raw Dataset that already exists in the file
    new_data : np.ndarray
        New data that we are writing
    data_hash : bytes
        Hash of the new data chunk
    hashed_slice : Slice
        Slice that is stored in the hash table for the given data_hash. This is a slice
        into the raw_data for the dataset; however if the data has not yet been written
        it may not point to a valid region in raw_data (but in that case it _would_
        point to a slice in ``slices_to_write``)
    chunk_being_written : np.ndarray
        New data chunk to be written
    slices_to_write : dict[slice, tuple] | None
        Dict of slices which will be written. Maps slices that will exist in the
        raw_data once the write is complete to slices of the dataset that is being
        written.
    data_to_write : dict[slice, tuple] | None
        Dict of arrays which will be written as chunks. Maps slices that will exist in
        the raw_data once the write is complete to chunks of the dataset that is being
        written. If ``data_to_write`` is specified, ``slices_to_write`` must be None.
    """
    if slices_to_write is not None and hashed_slice in slices_to_write:
        # The hash table contains a slice we will write but haven't yet; grab the
        # chunk from the new data being written
        reused_chunk = new_data[slices_to_write[hashed_slice].raw]
    elif data_to_write is not None and hashed_slice in data_to_write:
        # The hash table contains a slice we will write but haven't yet; grab the
        # chunk from the data_to_write dict, which stores the data that will be written
        # for the given hashed slice.
        reused_chunk = data_to_write[hashed_slice]
    else:
        # The hash table contains a slice that was written in a previous
        # write operation; grab that chunk from the existing raw data
        reused_slice = Tuple(
            hashed_slice, *[slice(0, size) for size in chunk_being_written.shape[1:]]
        )
        reused_chunk = raw_dataset[reused_slice.raw]

    # In some cases type coercion can happen during the write process even if the dtypes
    # are the same - for example, if the raw_data.dtype == dtype('O'), but the elements
    # are bytes, and chunk_being_written.dtype == dtype('O'), but the elements are
    # utf-8 strings. For this case, when the raw_data is changed, e.g.
    #      raw_data[some_slice] = chunk_being_written[another_slice]
    # the data that gets written is bytes. So in certain cases, just calling
    # assert_array_equal doesn't work. Instead, we convert each element to a bytestring
    # first.
    if reused_chunk.dtype == "O" and chunk_being_written.dtype == "O":
        to_be_written = _convert_to_bytes(chunk_being_written)
        to_be_reused = _convert_to_bytes(reused_chunk)
    else:
        to_be_written = chunk_being_written
        to_be_reused = reused_chunk

    try:
        assert_array_equal(to_be_reused, to_be_written)
    except AssertionError as e:
        raise ValueError(
            f"Hash {data_hash!r} of existing data chunk {reused_chunk} "
            f"matches the hash of new data chunk {chunk_being_written}, "
            "but data does not."
        ) from e


def _convert_to_bytes(arr: np.ndarray) -> np.ndarray:
    """Convert each element in the array to bytes.

    Each element in the array is assumed to be the same type, even if the input is an
    object dtype array.

    Parameters
    ----------
    arr : np.ndarray
        Array to be converted; no conversion is done if the elements are already bytes

    Returns
    -------
    np.ndarray
        Object dtype array filled with elements of type bytes
    """
    # Check the actual type of the first element of the array; if it's bytes,
    # there's no need to cast (and casting won't work anyway)
    if len(arr) > 0 and isinstance(arr.flatten()[0], bytes):
        return arr
    else:
        return np.vectorize(lambda i: bytes(i, encoding="utf-8"))(arr)


def write_dataset_chunks(f, name, data_dict):
    """
    data_dict should be a dictionary mapping chunk_size index to either an
    array for that chunk, or a slice into the raw data for that chunk

    """
    if name not in f["_version_data"]:
        raise NotImplementedError(
            "Use write_dataset() if the dataset does not yet exist"
        )

    raw_data = f["_version_data"][name]["raw_data"]
    chunks = tuple(raw_data.attrs["chunks"])
    chunk_size = chunks[0]

    validate_reused_chunks = os.environ.get(
        "ENABLE_CHUNK_REUSE_VALIDATION", "false"
    ).lower() in ("1", "true")

    with Hashtable(f, name) as hashtable:
        old_chunks = hashtable.largest_index
        chunks_reused = 0

        slices = {i: None for i in data_dict}

        # Mapping from slices in the raw dataset after this write is complete to ndarray
        # chunks of the new data which will be written
        data_to_write = {}
        for chunk, data_s in data_dict.items():
            if isinstance(data_s, (slice, tuple, Tuple, Slice)):
                slices[chunk] = ndindex(data_s)
            else:
                if data_s.dtype != raw_data.dtype:
                    raise ValueError(
                        f"dtypes do not match ({data_s.dtype} != {raw_data.dtype})"
                    )

                data_hash = hashtable.hash(data_s)

                if data_hash in hashtable:
                    hashed_slice = hashtable[data_hash]
                    slices[chunk] = hashed_slice

                    if validate_reused_chunks:
                        _verify_new_chunk_reuse(
                            raw_dataset=raw_data,
                            new_data=data_s,
                            data_hash=data_hash,
                            hashed_slice=hashed_slice,
                            chunk_being_written=data_s,
                            data_to_write=data_to_write,
                        )

                    chunks_reused += 1

                else:
                    idx = hashtable.largest_index
                    raw_slice = Slice(
                        idx * chunk_size, idx * chunk_size + data_s.shape[0]
                    )
                    slices[chunk] = raw_slice
                    hashtable[data_hash] = raw_slice
                    data_to_write[raw_slice] = data_s

        new_chunks = hashtable.largest_index

    assert None not in slices.values()
    old_shape = raw_data.shape
    raw_data.resize((old_shape[0] + len(data_to_write) * chunk_size,) + chunks[1:])
    for raw_slice, data_s in data_to_write.items():
        c = (raw_slice.raw,) + tuple(slice(0, i) for i in data_s.shape[1:])
        raw_data[c] = data_s

    logging.debug(
        "  %s: " "New chunks written: %d; " "Number of chunks reused: %d",
        name,
        new_chunks - old_chunks,
        chunks_reused,
    )

    return slices


def create_virtual_dataset(
    f, version_name, name, shape, slices, attrs=None, fillvalue=None
):
    from h5py._hl.selections import select
    from h5py._hl.vds import VDSmap

    raw_data = f["_version_data"][name]["raw_data"]
    raw_data_shape = raw_data.shape
    slices = {c: s.reduce() for c, s in slices.items()}

    if len(raw_data) == 0:
        layout = VirtualLayout(shape=(0,), dtype=raw_data.dtype)
    else:
        # Chunks in the raw dataset are expanded along the first dimension only.
        # Since the chunks are pointed to by virtual datasets, it doesn't make
        # sense to expand the chunks in the raw dataset along multiple dimensions
        # (the true layout of the chunks in the raw dataset is irrelevant).
        for c, s in slices.items():
            if len(c.args[0]) != len(s):
                raise ValueError(f"Inconsistent slices dictionary ({c.args[0]}, {s})")

        # h5py 3.3 changed the VirtualLayout code so that it no longer uses
        # sources. See https://github.com/h5py/h5py/pull/1905.
        layout = VirtualLayout(shape, dtype=raw_data.dtype)
        layout_has_sources = hasattr(layout, "sources")
        if not layout_has_sources:
            from h5py import _selector

            layout._src_filenames.add(b".")
            space = h5s.create_simple(shape)
            selector = _selector.Selector(space)

        for c, s in slices.items():
            if c.isempty():
                continue
            # idx = Tuple(s, *Tuple(*[slice(0, i) for i in shape[1:]]).as_subindex(Tuple(*c.args[1:])).args)
            S = [Slice(0, len(c.args[i])) for i in range(1, len(shape))]
            idx = Tuple(s, *S)
            # assert c.newshape(shape) == vs[idx.raw].shape, (c, shape, s)

            # This is equivalent to
            #
            # layout[c.raw] = vs[idx.raw]
            #
            # but faster because vs[idx.raw] does a deepcopy(vs), which is
            # slow. We need different versions for h5py 2 and 3 because the
            # virtual sources code was rewritten.
            if not layout_has_sources:
                key = idx.raw
                vs_sel = select(raw_data.shape, key, dataset=None)

                sel = selector.make_selection(c.raw)
                layout.dcpl.set_virtual(
                    sel.id, b".", raw_data.name.encode("utf-8"), vs_sel.id
                )

            else:
                vs_sel = select(raw_data_shape, idx.raw, None)
                layout_sel = select(shape, c.raw, None)
                layout.sources.append(
                    VDSmap(layout_sel.id, ".", raw_data.name, vs_sel.id)
                )

    dtype = raw_data.dtype
    if dtype.metadata and (
        "vlen" in dtype.metadata or "h5py_encoding" in dtype.metadata
    ):
        # Variable length string dtype
        # (https://h5py.readthedocs.io/en/2.10.0/strings.html). Setting the
        # fillvalue in this case doesn't work
        # (https://github.com/h5py/h5py/issues/941).
        if fillvalue not in [0, "", b"", None]:
            raise ValueError(
                "Non-default fillvalue not supported for variable length strings"
            )
        fillvalue = None

    virtual_data = f["_version_data/versions"][version_name].create_virtual_dataset(
        name, layout, fillvalue=fillvalue
    )

    if attrs:
        for k, v in attrs.items():
            virtual_data.attrs[k] = v
    virtual_data.attrs["raw_data"] = raw_data.name
    virtual_data.attrs["chunks"] = raw_data.chunks
    return virtual_data


def get_available_filters() -> Iterator[int]:
    """Retrieve all of the registered h5py filters.

    Returns
    -------
    Iterator[int]
        Filter ID numbers; each filter has a dedicated ID - see
        the docs for the particular filter being used for more information
        about these
    """
    for i in range(65536):
        if h5z.filter_avail(i):
            yield i



import hashlib
import struct
from collections.abc import MutableMapping
from functools import lru_cache

import numpy as np
from h5py import File
from ndindex import ChunkSize, Slice, Tuple

from .slicetools import spaceid_to_slice


class Hashtable(MutableMapping):
    """
    A proxy class representing the hash table for an array

    The hash table for an array is a mapping from {sha256_hash: slice}, where
    slice is a slice for the data in the array.

    General usage should look like

        with Hashtable(f, name) as h:
            data_hash = h.hash(data[raw_slice])
            raw_slice = h.setdefault(data_hash, raw_slice)

    where setdefault will insert the hash into the table if it
    doesn't exist, and return the existing entry otherwise.

    hashtable.largest_index is the next index in the array that slices
    should be mapped to.

    Note that for performance reasons, the hashtable does not write to the
    dataset until you call write() or it exit as a context manager.

    """

    hash_function = hashlib.sha256
    hash_size = hash_function().digest_size

    # Cache instances of the class for performance purposes. This works off
    # the assumption that nothing else modifies the version data.

    # This is done here because putting @lru_cache() on the class breaks the
    # classmethods. Warning: This does not normalize kwargs, so it is possible
    # to have multiple hashtable instances for the same hashtable.
    @lru_cache()  # type: ignore
    def __new__(cls, *args, **kwargs):
        obj = super().__new__(cls)
        return obj

    def __init__(self, f, name, *, chunk_size=None, hash_table_name="hash_table"):
        from .backend import DEFAULT_CHUNK_SIZE

        self.f = f
        self.name = name
        self.chunk_size = chunk_size or DEFAULT_CHUNK_SIZE
        self.hash_table_name = hash_table_name

        if hash_table_name in f["_version_data"][name]:
            self.hash_table_dataset = f["_version_data"][name][hash_table_name]
            self.hash_table, self._indices = self._load_hashtable(
                self.hash_table_dataset
            )
        else:
            self.hash_table_dataset = self._create_hashtable()
            self.hash_table = self.hash_table_dataset[:]
            self._indices = {}

        self._largest_index = None

    @classmethod
    def from_versions_traverse(
        cls,
        f: File,
        name: str,
    ):
        """Traverse all versions of a dataset, writing to a brand new hash table.

        Parameters
        ----------
        f : File
            File for which a hash table is to be generated
        name : str
            Name of the dataset for which a hash table is to be generated
        """
        name = name.removeprefix("/_version_data/")

        for version_group in f["_version_data"]["versions"].values():
            if name in version_group:
                dataset = version_group[name]
                with Hashtable(f, name) as hashtable:
                    if dataset.is_virtual:
                        for vs in dataset.virtual_sources():
                            sl = spaceid_to_slice(vs.src_space)
                            dl = spaceid_to_slice(vs.vspace)
                            assert len(dl.raw) == len(sl.raw)
                            data_slice = dataset[dl.raw]
                            slice_hash = hashtable.hash(data_slice)
                            if slice_hash not in hashtable:
                                hashtable[slice_hash] = sl.raw[0]

    @classmethod
    def from_raw_data(cls, f, name, chunk_size=None, hash_table_name="hash_table"):
        if hash_table_name in f["_version_data"][name]:
            raise ValueError(
                f"a hash table {hash_table_name!r} for {name!r} already exists"
            )

        hashtable = cls(f, name, chunk_size=chunk_size, hash_table_name=hash_table_name)

        raw_data = f["_version_data"][name]["raw_data"]
        chunks = ChunkSize(raw_data.chunks)
        for c in chunks.indices(raw_data.shape):
            data_hash = hashtable.hash(raw_data[c.raw])
            hashtable.setdefault(data_hash, c.args[0])

        hashtable.write()
        return hashtable

    def hash(self, data: np.ndarray):
        """
        Compute hash for `data` array.
        """
        # Object dtype arrays store the ids of the elements, which may or may not be
        # reused, making it unsuitable for hashing. Instead, we need to make a combined
        # hash with the value of each element.
        hash_value = self.hash_function()  # type: ignore
        if data.dtype == object:
            for value in data.flat:
                if isinstance(value, str):
                    # default to utf-8 encoding since it's a superset of ascii (the only
                    # other valid encoding supported in h5py)
                    value = value.encode("utf-8")
                # hash the length of value ('Q' is unsigned long long, which is 64 bit
                # on all of today's architectures)
                # Use little-endian byte order (e.g. x86, ARM) for consistency
                # everywhere, even on big-endian architectures (e.g. PowerPC).
                hash_value.update(struct.pack("<Q", len(value)))
                # Hash the buffer of bytes, bytearray, memoryview, etc.
                hash_value.update(value)
        else:
            hash_value.update(np.ascontiguousarray(data))

        hash_value.update(str(data.shape).encode("ascii"))
        return hash_value.digest()

    @property
    def largest_index(self):
        if self._largest_index is None:
            self._largest_index = self.hash_table_dataset.attrs["largest_index"]
        return self._largest_index

    @largest_index.setter
    def largest_index(self, value):
        self._largest_index = value

    def write(self):
        largest_index = self.largest_index
        if largest_index >= self.hash_table_dataset.shape[0]:
            self.hash_table_dataset.resize((largest_index,))

        # largest_index is here for backwards compatibility for when the hash
        # table shape used to always be chunk_size aligned.
        self.hash_table_dataset.attrs["largest_index"] = self.largest_index
        self.hash_table_dataset[:largest_index] = self.hash_table[:largest_index]

    def inverse(self):
        r"""
        Return a dictionary mapping Slice: array_of_hash.

        The Slices are all `reduce()`\d.
        """
        return {Slice(*s).reduce(): h for h, s in self.hash_table}

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        if exc_type is not None:
            return
        self.write()

    def _create_hashtable(self):
        f = self.f
        name = self.name

        # TODO: Use get_chunks() here (the real chunk size should be based on
        # bytes, not number of elements)
        dtype = np.dtype([("hash", "B", (self.hash_size,)), ("shape", "i8", (2,))])
        hash_table = f["_version_data"][name].create_dataset(
            self.hash_table_name,
            shape=(1,),
            dtype=dtype,
            chunks=(self.chunk_size,),
            maxshape=(None,),
            compression="lzf",
        )
        hash_table.attrs["largest_index"] = 0
        return hash_table

    def _load_hashtable(self, hash_table_dataset):
        largest_index = hash_table_dataset.attrs["largest_index"]
        hash_table_arr = hash_table_dataset[:largest_index]
        hashes = bytes(hash_table_arr["hash"])
        hashes = [
            hashes[i * self.hash_size : (i + 1) * self.hash_size]
            for i in range(largest_index)
        ]
        return hash_table_arr, {k: i for i, k in enumerate(hashes)}

    def __getitem__(self, key):
        if isinstance(key, np.ndarray):
            key = key.tobytes()
        i = self._indices[key]
        shapes = self.hash_table["shape"]
        return Slice(*shapes[i])

    def __setitem__(self, key, value):
        if isinstance(key, np.ndarray):
            key = key.tobytes()
        if not isinstance(key, bytes):
            raise TypeError(f"key must be bytes, got {type(key)}")
        if len(key) != self.hash_size:
            raise ValueError("key must be %d bytes" % self.hash_size)
        if isinstance(value, Tuple):
            if len(value.args) > 1:
                raise NotImplementedError(
                    "Chunking in more other than the first dimension"
                )
            value = value.args[0]
        if isinstance(value, slice):
            value = Slice(value.start, value.stop, value.step)
        elif isinstance(value, Slice):
            pass
        else:
            raise TypeError("value must be a slice object")
        if value.isempty():
            return
        if value.step not in [1, None]:
            raise ValueError("only step-1 slices are supported")

        kv = (list(key), (value.start, value.stop))
        if key in self._indices:
            if bytes(self.hash_table[self._indices[key]])[0] != key:
                raise ValueError(
                    "The key %s is already in the hashtable under another index."
                )
            self.hash_table[self._indices[key]] = kv
        else:
            if self.largest_index >= self.hash_table.shape[0]:
                newshape = (self.hash_table.shape[0] + self.chunk_size,)
                new_hash_table = np.zeros(newshape, dtype=self.hash_table.dtype)
                new_hash_table[: self.hash_table.shape[0]] = self.hash_table
                self.hash_table = new_hash_table
            self.hash_table[self.largest_index] = kv
            self._indices[key] = self.largest_index
            self.largest_index += 1

    def __delitem__(self, key):
        raise NotImplementedError

    def __iter__(self):
        return iter(self._indices)

    def __len__(self):
        return len(self._indices)



# This file allows cimport'ing the functions declared below from other Cython modules

from .cytools cimport hsize_t


cdef class IndexChunkMapper:
    cdef readonly hsize_t[:] chunk_indices
    cdef readonly hsize_t chunk_size
    cdef readonly hsize_t dset_size
    cdef readonly hsize_t n_chunks
    cdef readonly hsize_t last_chunk_size

    cpdef tuple[object, object, object] chunk_submap(self, hsize_t chunk_idx)
    cpdef tuple[object, object | None] read_many_slices_params(self)

    cpdef object chunks_indexer(self)
    cpdef object whole_chunks_idxidx(self)

    # Private methods. Cython complains if we don't export _everything_.
    cdef tuple[hsize_t, hsize_t] _chunk_start_stop(
        self, hsize_t chunk_idx
    ) noexcept nogil


cpdef hsize_t[:, :, :] read_many_slices_params_nd(
    transfer_type,  # TransferType Python enum
    list[IndexChunkMapper] mappers,
    hsize_t[:, :] chunk_idxidx,
    hsize_t[:] src_slab_offsets,
    hsize_t[:] dst_slab_offsets,
)


# At the moment of writing, Cython enums are unusable in pure python mode:
# https://github.com/cython/cython/issues/4252
# The below is for user reference only and not actually used in the code.

cdef enum ReadManySlicesColumn:
    """Axis 1 of the return value of IndexChunkMapper.read_many_slices_params"""
    chunk_sub_start = 0
    value_sub_start = 1
    count = 2
    chunk_sub_stride = 3
    value_sub_stride = 4


cdef enum ReadManySlicesNDColumn:
    """Axis 1 of the return value of read_many_slices_params_nd"""
    src_start = 0
    dst_start = 1
    count_ = 2  # Two enums can't define the same label
    src_stride = 3
    dst_stride = 4



from __future__ import annotations

from typing import Any

import ndindex
import numpy as np


def asarray(a, dtype=None):
    """Variant of np.asarray(a, dtype=dtype), with two differences:

    1. If a is a numpy-like array, don't coerce it to a numpy.ndarray
    2. If a has a ABI-compatible dtype, return a view instead of a copy
       (works around https://github.com/numpy/numpy/issues/27509)
    """
    if not hasattr(a, "__array__") or np.isscalar(a):
        return np.asarray(a, dtype=dtype)

    if dtype is None:
        return a

    dtype = np.dtype(dtype)
    if a.dtype == dtype:
        return a

    if (
        dtype.itemsize == a.itemsize
        and dtype.kind in ("i", "u")
        and a.dtype.kind in ("i", "u")
        and hasattr(a, "view")
    ):
        # Note that this does not reduce the amount of safety checks:
        # np.array(-1).astype("u1") doesn't raise and returns 255!
        return a.view(dtype)

    return a.astype(dtype)


def ix_with_slices(*idx: Any, shape: tuple[int, ...]) -> tuple:
    """Variant of numpy.ix_ with added support for mixing with slices.

    Given a numpy ndindex where each element could be either a 1D fancy integer
    array index or a slice, convert it to an index that numpy understands as 'for
    each axis, take these indices'.

    Scalar indices, None and np.newaxis are NOT supported.

    Example
    -------
    >>> a = np.arange(3 * 3 * 2).reshape(3, 3, 2)
    >>> a
    array([[[ 0,  1],
            [ 2,  3],
            [ 4,  5]],
           [[ 6,  7],
            [ 8,  9],
            [10, 11]],
           [[12, 13],
            [14, 15],
            [16, 17]]])
    >>> # Select planes 1 and 2, then the first 2 rows, then columns 1 and 0
    >>> a[[1, 2], :, :][: ,:2, :][:, :, [1, 0]]
    array([[[ 7,  6],
            [ 9,  8]],
           [[13, 12],
            [15, 14]]])
    >>> a[np.ix_([1, 2], [0, 1], [1, 0])]
    array([[[ 7,  6],
            [ 9,  8]],
           [[13, 12],
            [15, 14]]])
    >>> a[ix_with_slices([1, 2], slice(2), [1, 0], shape=a.shape)]
    array([[[ 7,  6],
            [ 9,  8]],
           [[13, 12],
            [15, 14]]])
    """
    if len(idx) < 2:
        return idx

    # A single fancy index can be mixed with slices without problems
    n_fancy = 0
    for i in idx:
        if not isinstance(i, slice):
            if i is None or i is np.newaxis or np.isscalar(i):
                raise NotImplementedError("Scalars, None and newaxis are not supported")
            n_fancy += 1
    if n_fancy < 2:
        return idx

    idx2 = []
    for axis, i in enumerate(idx):
        if isinstance(i, slice):
            i = ndindex.Slice(i).reduce(shape[axis])
            i = np.arange(i.start, max(-1, i.stop), i.step)
        idx2.append(i)

    return np.ix_(*idx2)


def format_ndindex(idx: Any) -> str:
    """Format a numpy ndindex for pretty-printing.

    >>> format_ndindex(slice(None))
    ':'
    >>> format_ndindex(slice(None, 10, 2))
    ':10:2'
    >>> format_ndindex(())
    '()'
    >>> format_ndindex((1, slice(2, 3, 1), np.array([4, 5, 6])))
    '1, 2:3, [4, 5, 6]'
    """
    if isinstance(idx, tuple):
        if idx == ():
            return "()"
    else:
        idx = (idx,)

    idx_s = []
    for i in idx:
        if isinstance(i, slice):
            start = "" if i.start is None else i.start
            stop = "" if i.stop is None else i.stop
            step = "" if i.step in (1, None) else f":{i.step}"
            idx_s.append(f"{start}:{stop}{step}")
        elif isinstance(i, np.ndarray):
            idx_s.append(str(i.tolist()))
        else:
            idx_s.append(str(i))
    return ", ".join(idx_s)



import math

from hypothesis import given
from hypothesis import strategies as st

from ..cytools import ceil_a_over_b, count2stop, smallest_step_after, stop2count


def free_slices_st(size: int):
    """Hypothesis draw of a slice object to slice an array of up to size elements"""
    start_st = st.integers(0, size)
    stop_st = st.integers(0, size)
    # only non-negative steps are allowed
    step_st = st.integers(1, size)
    return st.builds(slice, start_st, stop_st, step_st)


@given(free_slices_st(5))
def test_stop2count_count2stop(s):
    count = stop2count(s.start, s.stop, s.step)
    assert count == len(range(s.start, s.stop, s.step))

    stop = count2stop(s.start, count, s.step)
    # When count == 0 or when step>1, multiple stops yield the same count,
    # so stop won't necessarily be equal to s.stop
    assert count == len(range(s.start, stop, s.step))


@given(st.integers(0, 10), st.integers(1, 10))
def test_ceil_a_over_b(a, b):
    expect = math.ceil(a / b)
    actual = ceil_a_over_b(a, b)
    assert actual == expect


@given(st.data(), st.integers(0, 10), st.integers(1, 4))
def test_smallest_step_after(data, x, m):
    a = data.draw(st.integers(0, x))
    expect = a
    while expect < x:
        expect += m
    actual = smallest_step_after(x, a, m)
    assert actual == expect
